
(* Principal types/signatures for terms/modules.  Principal kinds for constructors are in SubjectReduction. *)

Require Import Tactics.
Require Import Sequence.
Require Import Static.
Require Import TSubst.
Require Import Formation.
Require Import Regularity.
Require Import Declarative.
Require Import Singleton.
Require Import AlgorithmEq.
Require Import SubjectReduction.
Require Import SoundEq.
Require Import CompleteEq.


Definition clinh (A : (* cl *) term) : term :=
  (match A with
   | cl_cn k => inh k
   | _ => nonsense
   end).


Fixpoint cinh (G : context) {struct G} : sub :=
  (match G with
   | nil => id
   | G'; A => dot (clinh (subst (cinh G') A)) (cinh G')
   end).


Lemma cinh_subof :
  forall G1 G2,
    wfh G1 G2
    -> subof sof G1 (cinh G2) (G1;; G2).
Proof.
intros G1 G2 Hwf.
induct Hwf.

(* nil *)
{
intros G.
simpl.
apply subof_id.
}

(* cn *)
{
intros G1 G2 k _ IH Hofk.
rewrite <- app_comm_cons.
simpl cinh.
simp_sub.
apply subof_dot; auto.
simp_sub.
apply sof_cn.
apply inh_ofc.
eapply subst_ofk; eauto.
}

(* tm *)
{
intros G1 G2 t _ IH _.
rewrite <- app_comm_cons.
simpl cinh.
apply subof_dot; auto.
simp_sub.
apply sof_tm.
}

(* md *)
{
intros G1 G2 k T _ IH Hof Hfst.
autorewrite with canonlist.
simpl cinh.
simp_sub.
apply subof_dot.
  {
  apply subof_dot; auto.
  simp_sub.
  apply sof_cn.
  apply inh_ofc.
  eapply subst_ofk; eauto.
  eapply ofs_fst; eauto.
  }
simp_sub.
apply sof_md.
}
Qed.


Lemma cinh_trunc :
  forall G,
    compose (sh (length G)) (cinh G) = id.
Proof.
intro G.
induct G; auto.
Qed.


Lemma shift_under_clinh :
  forall i j A, iscl A -> subst (under i (sh j)) (clinh A) = clinh (subst (under i (sh j)) A).
Proof.
intros i j A HA.
induct HA; auto.
intros.
simpl clinh.
simp_sub.
rewrite -> shift_under_inh.
reflexivity.
Qed.


Lemma compose_under_cinh :
  forall G i j,
    Forall iscl G
    -> compose (under (length G + i) (sh j)) (cinh (subst_ctx (under i (sh j)) G)) = compose (cinh G) (under i (sh j)).
Proof.
intros G i j HG.
induct HG.

(* nil *)
{
simpl length.
simpl cinh.
simp_sub.
reflexivity.
}

(* cons *)
{
intros A G HA _ IH.
simpl length.
replace (S (length G) + i) with (S (length G + i)) by omega.
rewrite -> under_succ.
simpl cinh.
simp_sub.
rewrite -> shift_under_clinh; auto using subst_cl.
rewrite <- under_plus.
rewrite -> IH.
simp_sub.
reflexivity.
}
Qed.


Inductive lclose : (* ps *) term -> (* ps *) term -> (* ps *) term -> Prop :=
| lclose_down {T P}
    : lclose (ps_down T) P P

| lclose_ex {k P1 P2 P2'}
    : lclose P1 P2 P2'
      -> lclose (ps_ex k P1) P2 (ps_ex k P2')
.


Lemma subst_lclose :
  forall s P Q R G T,
    lclose P Q R
    -> linv P G T
    -> lclose (subst s P) (subst (under (length G) s) Q) (subst s R).
Proof.
intros s P Q R G T Hclose Hinv.
revert s G T Hinv.
induct Hclose.

(* down *)
{
intros T P s G Tinv Hinv.
invert Hinv; [].
intros <- _.
clear Tinv Hinv.
simp_sub.
apply lclose_down.
}

(* ex *)
{
intros k P1 P2 P2' _ IH s GG T Hinv.
invert Hinv; [].
intros G Hinv' <-.
so (IH (under 1 s) _ _ Hinv') as Hclose.
rewrite -> app_length.
simpl length.
simp_sub.
apply lclose_ex.
rewrite <- under_plus in Hclose.
rewrite -> under_succ in Hclose.
rewrite -> under_zero in Hclose.
simp_sub in Hclose.
exact Hclose.
}
Qed.


(* There is a discrepancy between this definition and the definition of
   principal kinds: in checking contexts, the principal kind judgement
   just uses ofc, while here we synthesize a principal type/signature and
   check equality/subsignature.  This is because we want to show that
   print/prins implies oftp/ofmp.
*)
Inductive print : context -> (* tm *) term -> (* cn *) term -> Prop :=
| print_var {G i t}
    : index i G (cl_tm t)
      -> print G (var i) (subst (sh (S i)) t)

| print_star {G}
    : print G tm_star cn_unit

| print_lam {G t1 t2 e}
    : ofc G t1 kd_type
      -> print (G; cl_tm t1) e t2
      -> print G (tm_lam t1 e) (cn_arrow t1 (subst1 nonsense t2))

| print_app {G t t1 t2 t1' e1 e2}
    : print G e1 t
      -> headnorm G t (cn_arrow t1 t2)
      -> print G e2 t1'
      -> eqc G t1' t1 kd_type
      -> print G (tm_app e1 e2) t2

| print_pair {G t1 t2 e1 e2}
    : print G e1 t1
      -> print G e2 t2
      -> print G (tm_pair e1 e2) (cn_prod t1 t2)

| print_pi1 {G t t1 t2 e}
    : print G e t
      -> headnorm G t (cn_prod t1 t2)
      -> print G (tm_pi1 e) t1

| print_pi2 {G t t1 t2 e}
    : print G e t
      -> headnorm G t (cn_prod t1 t2)
      -> print G (tm_pi2 e) t2

| print_plam {G k t e}
    : ofk G k
      -> print (G; cl_cn k) e t
      -> svalue e
      -> print G (tm_plam k e) (cn_all k t)

| print_papp {G u k t e c}
    : print G e u
      -> headnorm G u (cn_all k t)
      -> ofc G c k
      -> print G (tm_papp e c) (subst1 c t)

| print_pack {G k t c e u}
    : ofc G c k
      -> print G e u
      -> eqc G u (subst1 c t) kd_type
      -> ofc (G; cl_cn k) t kd_type
      -> print G (tm_pack c e k t) (cn_exists k t)

| print_unpack {G u k t t' e1 e2}
    : print G e1 u
      -> headnorm G u (cn_exists k t)
      -> print (G; cl_cn k; cl_tm t) e2 t'
      -> eqc (G; cl_cn k) (subst (dot nonsense id) t') (subst (dot nonsense (dot (shift1 (inh k)) sh1)) t') kd_type
      -> print G (tm_unpack e1 e2) (subst (dot nonsense (dot (inh k) id)) t')

| print_fix {G t e u}
    : print G e u
      -> eqc G u (cn_arrow (cn_arrow cn_unit t) t) kd_type
      -> print G (tm_fix t e) t

| print_lett {G t1 t2 e1 e2}
    : print G e1 t1
      -> print (G; cl_tm t1) e2 t2
      -> print G (tm_lett e1 e2) (subst1 nonsense t2)

| print_letm {G P G' T k t M e}
    : prinps G M P
      -> linv P G' T
      -> fsts T k
      -> print (G;; G'; cl_cn k; cl_md (shift1 T) (var 0)) (subst (under 2 (sh (length G'))) e) t
      -> eqc (G;; G'; cl_cn k) (subst (dot nonsense id) t) (subst (dot nonsense (dot (subst (compose (cinh G') (sh (S (length G')))) (inh k)) (compose (cinh G') (sh (S (length G')))))) t) kd_type
      -> print G (tm_letm M e) (subst (dot nonsense (dot (subst (cinh G') (inh k)) (cinh G'))) t)

| print_ext {G t M}
    : prins G M (sg_dyn t)
      -> print G (tm_ext M) t


with prins : context -> (* md *) term -> (* sg *) term -> Prop :=
| prins_var {G i T c}
    : index i G (cl_md T c)
      -> prins G (var i) (singles (subst (sh (S i)) c) (subst (sh (S i)) T))

| prins_star {G}
    : prins G md_star sg_one

| prins_stat {G c k}
    : prink G c k
      -> prins G (md_stat c) (sg_stat k)

| prins_dyn {G e t}
    : print G e t
      -> prins G (md_dyn e) (sg_dyn t)

| prins_lam {G T1 T2 k1 M}
    : ofs G T1
      -> fsts T1 k1
      -> prins (G; cl_cn k1; cl_md (shift1 T1) (var 0)) M (shift1 T2)
      -> prins G (md_lam T1 M) (sg_pi T1 T2)

| prins_app {G T1 T2 T1' M1 M2 c2}
    : prins G M1 (sg_pi T1 T2)
      -> prins G M2 T1'
      -> subs G T1' T1
      -> fstm G M2 c2
      -> prins G (md_app M1 M2) (subst1 c2 T2)

| prins_pair {G T1 T2 M1 M2}
    : prins G M1 T1
      -> prins G M2 T2
      -> prins G (md_pair M1 M2) (sg_sigma T1 (shift1 T2))

| prins_pi1 {G T1 T2 M}
    : prins G M (sg_sigma T1 T2)
      -> prins G (md_pi1 M) T1

| prins_pi2 {G T1 T2 M c}
    : prins G M (sg_sigma T1 T2)
      -> fstm G M c
      -> prins G (md_pi2 M) (subst1 (cn_pi1 c) T2)

| prins_lett {G t T e M}
    : print G e t
      -> prins (G; cl_tm t) M T
      -> prins G (md_lett e M) (subst1 nonsense T)

| prins_cir {G M P}
    : prinps G M P
      -> prins G (md_cir M) (sg_circ P)


with prinps : context -> (* lm *) term -> (* ps *) term -> Prop :=
| prinps_ret {G M T}
    : prins G M T
      -> prinps G (lm_ret M) (ps_down T)

| prinps_seal {G M T T'}
    : prins G M T'
      -> subs G T' T
      -> prinps G (lm_seal M T) (ps_down T)

| prinps_bind {G M1 M2 P1 G' T k P2 P2'}
    : prins G M1 (sg_circ P1)
      -> linv P1 G' T
      -> fsts T k
      -> prinps (G;; G'; cl_cn k; cl_md (shift1 T) (var 0)) (subst (under 2 (sh (length G'))) M2) P2
      -> lclose P1 (ps_ex k (subst1 nonsense P2)) P2'
      -> prinps G (lm_bind M1 M2) P2'

| prinps_unpack {G u k t P e M}
    : print G e u
      -> headnorm G u (cn_exists k t)
      -> prinps (G; cl_cn k; cl_tm t) M P
      -> prinps G (lm_unpack e M) (ps_ex k (subst1 nonsense P))
.


Scheme print_mut_ind := Minimality for print Sort Prop
with   prins_mut_ind := Minimality for prins Sort Prop
with   prinps_mut_ind := Minimality for prinps Sort Prop.
Combined Scheme prin_ind from print_mut_ind, prins_mut_ind, prinps_mut_ind.


Lemma weaken_prin_gen :
  (forall i j G1 G2 G3 e t,
     length G2 = i
     -> length G3 = j
     -> Forall iscl G3
     -> print (G1;; G3) e t
     -> print (G1;; G2;; subst_ctx (sh i) G3) (subst (under j (sh i)) e) (subst (under j (sh i)) t))
  /\
  (forall i j G1 G2 G3 M T,
     length G2 = i
     -> length G3 = j
     -> Forall iscl G3
     -> prins (G1;; G3) M T
     -> prins (G1;; G2;; subst_ctx (sh i) G3) (subst (under j (sh i)) M) (subst (under j (sh i)) T))
  /\
  (forall i j G1 G2 G3 M T,
     length G2 = i
     -> length G3 = j
     -> Forall iscl G3
     -> prinps (G1;; G3) M T
     -> prinps (G1;; G2;; subst_ctx (sh i) G3) (subst (under j (sh i)) M) (subst (under j (sh i)) T)).
Proof.
exploit
  (prin_ind
     (fun G e t =>
        forall i j G1 G2 G3,
          G = G1;; G3
          -> length G2 = i
          -> length G3 = j
          -> Forall iscl G3
          -> print (G1;; G2;; subst_ctx (sh i) G3) (subst (under j (sh i)) e) (subst (under j (sh i)) t))
     (fun G M T =>
        forall i j G1 G2 G3,
          G = G1;; G3
          -> length G2 = i
          -> length G3 = j
          -> Forall iscl G3
          -> prins (G1;; G2;; subst_ctx (sh i) G3) (subst (under j (sh i)) M) (subst (under j (sh i)) T))
     (fun G M T =>
        forall i j G1 G2 G3,
          G = G1;; G3
          -> length G2 = i
          -> length G3 = j
          -> Forall iscl G3
          -> prinps (G1;; G2;; subst_ctx (sh i) G3) (subst (under j (sh i)) M) (subst (under j (sh i)) T))) as Hind.

(* var *)
intros G n t Hindex i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
so (Compare_dec.lt_dec n j) as [Hlt | Hnlt].
  rewrite -> project_under_lt; auto; [].
  so (index_app_lt _#5 (eqconv (eqsymm Hlen3) Hlt) Hindex) as Hindex'.
  so (index_subst_ctx _#3 (sh i) Hindex') as Hindex''.
  rewrite -> Hlen3 in Hindex''.
  so (index_app_left _ _ (G1;; G2) _ _ Hindex'') as Hindex'''.
  simp_sub in Hindex'''.
  so (print_var Hindex''') as Hprint.
  simp_sub in Hprint.
  rewrite -> compose_sh_under_leq; [| omega].
  repl (j - (1 + n)) with (j - n - 1) by omega.
  assumption.

  assert (n >= j) as Hgeq by omega; clear Hnlt.
  rewrite -> project_under_ge; auto; [].
  simp_sub.
  repl (j + i + (n - j)) with (n + i) by omega.
  so (index_app_ge _#5 (eqconv (eqsymm Hlen3) Hgeq) Hindex) as Hindex'.
  so (index_app_right _ (G2;; subst_ctx (sh i) G3) _#3 Hindex') as Hindex''.
  rewrite -> app_length in Hindex''; [].
  rewrite -> subst_ctx_length in Hindex''; [].
  rewrite -> Hlen2 in Hindex''; [].
  rewrite -> Hlen3 in Hindex''; [].
  rewrite <- app_assoc in Hindex''; [].
  repl (n - j + (j + i)) with (n + i) in Hindex'' by omega.
  so (print_var Hindex'') as Hprint.
  rewrite -> compose_sh_under_geq; auto; [].
  simp_sub.
  repl (1 + n - j + i + j) with (S (n + i)) by omega.
  assumption.

(* star *)
intros G i j G1 G2 G3 -> H0 H1 H2.     
simp_sub.
apply print_star; done.

(* lam *)
intros G t1 t2 e Hoft1 _ IH i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
replsub (subst (dot nonsense (under j (sh i))) t2) with (subst1 nonsense (subst (under (S j) (sh i)) t2)).
#2 apply print_lam.
  eapply subst_ofc_type; eauto; [].
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.

  rewrite <- under_succ; [].
  rewrite -> app_comm_cons; [].
  repl (subst_ctx (sh i) G3; cl_tm (subst (under j (sh i)) t1)) with (subst_ctx (sh i) (G3; cl_tm t1)) by (simpl; simp_sub; subst; reflexivity).
  #2 apply IH; auto.
    simpl; auto; done.

    apply Forall_cons; auto with static; done.

(* app *)
intros G t t1 t2 t1' e1 e2 _ IH1 Hnorm _ IH2 Heqc i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
so (IH1 _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint1.
so (IH2 _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint2.
so (weaken_headnorm_gen _#7 Hlen2 Hlen3 Hnorm) as Hnorm'.
simp_sub in Hnorm'.
eapply print_app; eauto; [].
eapply subst_eqc_type; eauto; [].
subst i j.
apply subof_under_sof; auto; [].
apply subof_sh; [].
apply app_truncate; done.

(* pair *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
so (IH1 _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint1.
so (IH2 _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint2.
eapply print_pair; eauto; done.

(* pi1 *)
intros G t t1 t2 e _ IH Hnorm i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
so (IH _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint.
so (weaken_headnorm_gen _#7 Hlen2 Hlen3 Hnorm) as Hnorm'.
simp_sub in Hnorm'.
eapply print_pi1; eauto; done.

(* pi2 *)
intros G t t1 t2 e _ IH Hnorm i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
so (IH _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint.
so (weaken_headnorm_gen _#7 Hlen2 Hlen3 Hnorm) as Hnorm'.
simp_sub in Hnorm'.
eapply print_pi2; eauto; done.

(* plam *)
intros G k t e Hofk _ IH Hval i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
#3 apply print_plam.
  eapply subst_ofk; eauto; [].
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.

  rewrite <- under_succ; [].
  rewrite -> app_comm_cons; [].
  repl (subst_ctx (sh i) G3; cl_cn (subst (under j (sh i)) k)) with (subst_ctx (sh i) (G3; cl_cn k)) by (simpl; simp_sub; subst; reflexivity).
  #2 apply IH; auto.
    simpl; auto; done.

    apply Forall_cons; auto with static; done.

  apply subst_svalue; auto; done.

(* papp *)
intros G u k t e c _ IH Hnorm Hofc i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
so (IH _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint.
so (weaken_headnorm_gen _#7 Hlen2 Hlen3 Hnorm) as Hnorm'.
simp_sub in Hnorm'.
rewrite -> split_dot; rewrite -> subst_compose; [].
eapply print_papp; eauto; [].
eapply subst_ofc; eauto; [].
subst i j.
apply subof_under_sof; auto; [].
apply subof_sh; [].
apply app_truncate; done.

(* pack *)
intros G k t c e u Hofc _ IH Heqc Hoft i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
assert (subof sof (G1;; G2;; subst_ctx (sh i) G3) (under j (sh i)) (G1;; G3)) as Hsubof.
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.
#2 eapply print_pack; eauto using subst_ofc.
  simp_sub.
  so (subst_eqc _#6 Hsubof Heqc) as Heqc'.
  simp_sub in Heqc'.
  assumption.

  rewrite <- under_succ; [].
  rewrite -> app_comm_cons; [].
  repl (subst_ctx (sh i) G3; cl_cn (subst (under j (sh i)) k)) with (subst_ctx (sh i) (G3; cl_cn k)) by (simpl; simp_sub; subst; reflexivity).
  eapply subst_ofc_type; eauto; [].
  subst i j.
  repl (G1;; G3; cl_cn k) with (G1;; (G3; cl_cn k)) by (simpl; reflexivity).
  #2 apply subof_under_sof; auto.
    apply Forall_cons; auto with static; done.

    apply subof_sh; [].
    apply app_truncate; done.

(* unpack *)
intros G u k t t' e1 e2 _ IH1 Hnorm _ IH2 Heqc i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
so (IH1 _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint1.
so (weaken_headnorm_gen _#7 Hlen2 Hlen3 Hnorm) as Hnorm'.
simp_sub in Hnorm'.
rewrite -> shift_under_inh; [].
replsub (subst (dot nonsense (dot (inh (subst (under j (sh i)) k)) (under j (sh i)))) t') with (subst (dot nonsense (dot (inh (subst (under j (sh i)) k)) id)) (subst (under (S (S j)) (sh i)) t')).
#2 eapply print_unpack; eauto.
  simp_sub.
  rewrite <- !under_succ2; [].
  rewrite <- under_succ; [].
  rewrite -> app_comm_cons; [].
  rewrite -> app_comm_cons; [].
  repl (subst_ctx (sh i) G3; cl_cn (subst (under j (sh i)) k); cl_tm (subst (under (S j) (sh i)) t)) with (subst_ctx (sh i) (G3; cl_cn k; cl_tm t)) by (simpl; simp_sub; subst; reflexivity).
  #2 apply IH2; auto.
    simpl; auto; done.

    apply Forall_cons; auto with static; done.

  simp_sub. 
  #2 sor (refine (subst_eqc (G1;; G2;; subst_ctx (sh i) (G3; cl_cn k)) _ (under (S j) (sh i)) _#4 Heqc)) as Heqc'.
    subst i j.
    repl (G1;; G3; cl_cn k) with (G1;; (G3; cl_cn k)) by (simpl; reflexivity).
    #2 apply subof_under_sof; auto.
      apply Forall_cons; auto with static; done.
  
      apply subof_sh; [].
      apply app_truncate; done.
  simpl in Heqc'.
  simp_sub in Heqc'.
  replsub (subst (compose (under j (sh i)) sh1) (inh k)) with (subst (sh 1) (subst (under j (sh i)) (inh k))) in Heqc'.
  rewrite -> shift_under_inh in Heqc'.
  subst i j; assumption.

(* fix *)
intros G t e u _ IH Heqc i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
eapply print_fix; eauto; [].
replsub (cn_arrow (cn_arrow cn_unit (subst (under j (sh i)) t)) (subst (under j (sh i)) t)) with (subst (under j (sh i)) (cn_arrow (cn_arrow cn_unit t) t)).
eapply subst_eqc_type; eauto; [].
subst i j.
apply subof_under_sof; auto; [].
apply subof_sh; [].
apply app_truncate; done.

(* lett *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
replsub (subst (dot nonsense (under j (sh i))) t2) with (subst1 nonsense (subst (under (S j) (sh i)) t2)).
eapply print_lett; eauto; [].
rewrite <- under_succ; [].
rewrite -> app_comm_cons; [].
repl (subst_ctx (sh i) G3; cl_tm (subst (under j (sh i)) t1)) with (subst_ctx (sh i) (G3; cl_tm t1)) by (simpl; simp_sub; subst; reflexivity).
#2 apply IH2; auto.
  simpl; auto; done.

  apply Forall_cons; auto with static; done.

(* letm *)
{
intros GG P G T k t M e _ IH1 Hinv Hfst _ IH2 Heqc i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
so (linv_iscl _#3 Hinv) as HallclG.
simp_sub.
repl (subst (dot nonsense (dot (subst (compose (cinh G) (under j (sh i))) (inh k)) (compose (cinh G) (under j (sh i))))) t) with (subst (dot nonsense (dot (subst (cinh (subst_ctx (under j (sh i)) G)) (subst (under (length G + j) (sh i)) (inh k))) (cinh (subst_ctx (under j (sh i)) G)))) (subst (under (S (S (length G + j))) (sh i)) t)).
  {
  simp_sub.
  rewrite -> compose_under_cinh; auto.
  }
rewrite -> shift_under_inh.
eapply print_letm; eauto using subst_fsts.
  {
  rewrite -> under_plus.
  apply subst_linv; auto.
  }

  {
  simp_sub.
  replsub (subst (compose (under (length G + j) (sh i)) sh1) T) with (subst (under (S (length G + j)) (sh i)) (shift1 T)).
  rewrite -> app_assoc.
  rewrite -> app_comm_cons.
  rewrite -> app_comm_cons.
  repl (subst_ctx (sh i) G3;; subst_ctx (under j (sh i)) G; cl_cn (subst (under (length G + j) (sh i)) k); cl_md (subst (under (S (length G + j)) (sh i)) (shift1 T)) (var 0)) with (subst_ctx (sh i) (G3;; G; cl_cn k; cl_md (shift1 T) (var 0))).
    {
    simpl subst_ctx.
    rewrite -> subst_ctx_app.
    simp_sub.
    subst.
    rewrite -> app_length.
    reflexivity.
    }
  rewrite -> subst_ctx_length.
  exploit (IH2 i (S (S (length G + j))) G1 G2 (G3;; G; cl_cn k; cl_md (shift1 T) (var 0))) as H; auto.
    {
    autorewrite with canonlist.
    reflexivity.
    }

    {
    simpl length.
    rewrite -> app_length.
    omega.
    }

    {
    auto using iscl, Forall_app, Forall_cons.
    }
  force_exact H.
  simp_sub.
  f_equal.
  f_equal.
  f_equal.
  f_equal.
  simp_sub.
  rewrite <- compose_assoc.
  rewrite -> compose_sh_under_leq; [| omega].
  replace (length G + j - length G) with j by omega.
  simp_sub.
  reflexivity.
  }

  {
  exploit (subof_under_sof (G1;; G2) G1 (G3;; G; cl_cn k) (sh i)) as Hsubof.
    {
    eauto using Forall_app, Forall_cons, iscl.
    }

    {
    apply subof_sh.
    subst i.
    apply app_truncate.
    }
  autorewrite with canonlist in Hsubof.
  simpl length in Hsubof.
  rewrite -> app_length in Hsubof.
  so (subst_eqc _#6 Hsubof Heqc) as H.
  simpl subst_ctx in H.
  rewrite -> subst_ctx_app in H.
  autorewrite with canonlist in H.
  rewrite -> app_length in H.
  rewrite -> !Hlen3 in H.
  force_exact H.
  simp_sub.
  f_equal; [].
  f_equal; [].
  f_equal; [].
  f_equal.
    {
    rewrite -> subst_ctx_length.
    rewrite <- shift_under_inh.
    rewrite <- subst_compose.
    f_equal; [].
    setoid_rewrite <- compose_assoc at 3.
    rewrite -> compose_under_cinh; auto.
    rewrite -> compose_assoc.
    setoid_rewrite <- compose_assoc at 2.
    rewrite -> compose_sh_under_leq; [| omega].
    replace (length G + j - length G) with j by omega.
    simp_sub.
    reflexivity.
    }

    {
    rewrite -> subst_ctx_length.
    setoid_rewrite <- compose_assoc at 3.
    rewrite -> compose_under_cinh; auto.
    rewrite -> compose_assoc.
    setoid_rewrite <- compose_assoc at 2.
    rewrite -> compose_sh_under_leq; [| omega].
    replace (length G + j - length G) with j by omega.
    simp_sub.
    reflexivity.
    }
  }
}

(* ext *)
intros G t M _ IH i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
eapply print_ext; eauto; done.

(* var *)
intros G n T c Hindex i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
so (Compare_dec.lt_dec n j) as [Hlt | Hnlt].
  rewrite -> project_under_lt; auto; [].
  so (index_app_lt _#5 (eqconv (eqsymm Hlen3) Hlt) Hindex) as Hindex'.
  so (index_subst_ctx _#3 (sh i) Hindex') as Hindex''.
  rewrite -> Hlen3 in Hindex''.
  so (index_app_left _ _ (G1;; G2) _ _ Hindex'') as Hindex'''.
  simp_sub in Hindex'''.
  so (prins_var Hindex''') as Hprins.
  simp_sub in Hprins.
  rewrite -> shift_under_singles; [].
  simp_sub.
  rewrite -> compose_sh_under_leq; [| omega].
  repl (j - (1 + n)) with (j - n - 1) by omega.
  assumption.

  assert (n >= j) as Hgeq by omega; clear Hnlt.
  rewrite -> project_under_ge; auto; [].
  simp_sub.
  repl (j + i + (n - j)) with (n + i) by omega.
  so (index_app_ge _#5 (eqconv (eqsymm Hlen3) Hgeq) Hindex) as Hindex'.
  so (index_app_right _ (G2;; subst_ctx (sh i) G3) _#3 Hindex') as Hindex''.
  rewrite -> app_length in Hindex''; [].
  rewrite -> subst_ctx_length in Hindex''; [].
  rewrite -> Hlen2 in Hindex''; [].
  rewrite -> Hlen3 in Hindex''; [].
  rewrite <- app_assoc in Hindex''; [].
  repl (n - j + (j + i)) with (n + i) in Hindex'' by omega.
  so (prins_var Hindex'') as Hprins.
  rewrite -> shift_under_singles; [].
  simp_sub.
  rewrite -> compose_sh_under_geq; auto; [].
  simp_sub.
  repl (1 + n - j + i + j) with (S (n + i)) by omega.
  assumption.

(* star *)
intros G i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
apply prins_star; done.

(* stat *)
intros G c k Hprink i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
apply prins_stat; [].
apply weaken_prink_gen; auto; done.

(* dyn *)
intros G e t _ IH i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
apply prins_dyn; auto; done.

(* lam *)
intros G T1 T2 k1 M HofT1 Hfst _ IH i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
#2 eapply prins_lam; eauto using subst_fsts.
  eapply subst_ofs; eauto; [].
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.

  simp_sub.
  #2 exploit (IH i (S (S j)) G1 G2 (G3; cl_cn k1; cl_md (shift1 T1) (var 0))) as H; auto with static.
    simpl.
    omega.
  simpl in H.
  simp_sub in H.
  subst i j; eassumption.

(* app *)
intros G T1 T2 T1' M1 M2 c2 _ IH1 _ IH2 Hsubs Hfstm i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
rewrite -> split_dot; rewrite -> subst_compose; [].
so (IH1 _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprin1.
simp_sub in Hprin1.
#2 eapply prins_app; eauto.
  eapply subst_subs; eauto; [].
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.

  eapply subst_fstm; eauto; [].
  subst i j.
  apply subof_under_tof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.

(* pair *)
intros G T1 T2 M1 M2 _ IH1 _ IH2 i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
rewrite -> subst_compose; [].
eapply prins_pair; eauto; done.

(* pi1 *)
intros G T1 T2 M _ IH i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
eapply prins_pi1; eauto; done.

(* pi2 *)
intros G T1 T2 M c _ IH Hfstm i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
rewrite -> split_dot; rewrite -> subst_compose; [].
so (IH _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as H.
simp_sub in H.
eapply prins_pi2; eauto; [].
eapply subst_fstm; eauto; [].
subst i j.
apply subof_under_tof; auto; [].
apply subof_sh; [].
apply app_truncate; done.

(* lett *)
intros G t T e M _ IH1 _ IH2 i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
replsub (subst (dot nonsense (under j (sh i))) T) with (subst1 nonsense (subst (under (S j) (sh i)) T)).
eapply prins_lett; eauto; [].
rewrite <- under_succ; [].
rewrite -> app_comm_cons; [].
repl (subst_ctx (sh i) G3; cl_tm (subst (under j (sh i)) t)) with (subst_ctx (sh i) (G3; cl_tm t)) by (simpl; simp_sub; subst; reflexivity).
#2 apply IH2; auto.
  simpl; auto; done.

  apply Forall_cons; auto with static; done.

(* cir *)
{
intros G M P _ IH i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
apply prins_cir; auto.
}

(* ret *)
{
intros G M T _ IH i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
apply prinps_ret; auto.
}

(* seal *)
{
intros G M T T' _ IH Hsub i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
eapply prinps_seal; eauto.
eapply subst_subs; eauto; [].
subst i j.
apply subof_under_sof; auto; [].
apply subof_sh; [].
apply app_truncate; done.
}

(* bind *)
{
intros G M1 M2 P1 G' T k P2 P2' _ IH1 Hinv Hfst _ IH2 Hclose i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
apply (@prinps_bind _ _ _
         (subst (under j (sh i)) P1)
         (subst_ctx (under j (sh i)) G')
         (subst (under (j + length G') (sh i)) T)
         (subst (under (j + length G') (sh i)) k)
         (subst (under (2 + j + length G') (sh i)) P2)).
  {
  apply IH1; auto.
  }

  {
  so (subst_linv (under j (sh i)) _#3 Hinv) as Hinv'.
  rewrite <- under_plus in Hinv'.
  replace (j + length G') with (length G' + j) by omega.
  exact Hinv'.
  }

  {
  apply subst_fsts; auto.
  }

  {
  simp_sub.
  exploit (IH2 i (j + length G' + 2) G1 G2 (G3;; G'; cl_cn k; cl_md (shift1 T) (var 0))) as H.
    {
    autorewrite with canonlist.
    reflexivity.
    }

    { assumption. }

    {
    simpl length.
    rewrite -> app_length.
    omega.
    }

    {
    apply Forall_cons; auto using iscl.
    apply Forall_cons; auto using iscl.
    apply Forall_app; auto.
    eapply linv_iscl; eauto.
    }

    {
    force_exact H.
    f_equal.
      {
      rewrite -> !subst_ctx_cons.
      rewrite -> subst_ctx_app.
      simp_sub.
      rewrite -> !app_length.
      simpl length.
      rewrite -> !app_length.
      replace (j + length G') with (length G' + j) by omega.
      subst i j.
      simp_sub.
      rewrite <- !app_comm_cons.
      rewrite <- app_assoc.
      reflexivity.
      }

      {
      simp_sub.
      f_equal.
      rewrite -> project_under_lt; [| omega].
      rewrite -> project_compose.
      simp_sub.
      rewrite -> project_under_lt; [| omega].
      f_equal.
      f_equal.
      rewrite <- compose_assoc.
      unfold sh1.
      rewrite -> compose_sh_sh.
      rewrite -> compose_sh_under_leq; [| omega].
      rewrite -> subst_ctx_length.
      f_equal.
        {
        f_equal.
        omega.
        }

        {
        f_equal.
        omega.
        }
      }

      {
      simp_sub.
      reflexivity.
      }
    }
  }

  {
  so (subst_lclose (under j (sh i)) _#5 Hclose Hinv) as Hclose'.
  rewrite <- under_plus in Hclose'.
  simp_sub.
  simp_sub in Hclose'.
  replace (j + length G') with (length G' + j) by omega.
  replace (2 + (length G' + j)) with (S (S (length G' + j))) by omega.
  rewrite -> under_succ.
  simp_sub.
  exact Hclose'.
  }
}

(* unpack *)
{
intros G u k t P e M _ IH1 Hnorm _ IH2 i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
so (IH1 _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint1.
so (weaken_headnorm_gen _#7 Hlen2 Hlen3 Hnorm) as Hnorm'.
simp_sub in Hnorm'.
setoid_rewrite -> split_dot at 3.
rewrite -> subst_compose.
eapply prinps_unpack; eauto.
simp_sub.
rewrite <- !under_succ2; [].
rewrite <- under_succ; [].
rewrite -> app_comm_cons; [].
rewrite -> app_comm_cons; [].
repl (subst_ctx (sh i) G3; cl_cn (subst (under j (sh i)) k); cl_tm (subst (under (S j) (sh i)) t)) with (subst_ctx (sh i) (G3; cl_cn k; cl_tm t)) by (simpl; simp_sub; subst; reflexivity).
apply IH2; auto.
  {
  simpl.
  omega.
  }

  {
  apply Forall_cons; auto using iscl.
  }
}

(* epilogue *)
destruct_all Hind.
do2 2 split; intros; eauto; done.
Qed.


Lemma weaken1_prins :
  forall G M T A,
    prins G M T
    -> prins (G; A) (shift1 M) (shift1 T).
Proof.
intros G M T A Hprins.
so (weaken_prin_gen anderl 1 0 G (nil; A) nil _ _ (eq_refl _) (eq_refl _) (Forall_nil _) Hprins) as H.
simp_sub in H.
assumption.
Qed.


Lemma weaken1_prinps :
  forall G M T A,
    prinps G M T
    -> prinps (G; A) (shift1 M) (shift1 T).
Proof.
intros G M T A Hprins.
so (weaken_prin_gen anderr 1 0 G (nil; A) nil _ _ (eq_refl _) (eq_refl _) (Forall_nil _) Hprins) as H.
simp_sub in H.
assumption.
Qed.


Lemma linv_fun :
  forall P G G' T T',
    linv P G T
    -> linv P G' T'
    -> G = G' /\ T = T'.
Proof.
intros P G G' T T' Hinv Hinv'.
revert G' T' Hinv'.
induct Hinv.

(* down *)
{
intros T G' T' Hinv'.
invert Hinv'; [].
intros <- <-.
auto.
}

(* ex *)
{
intros k P G T _ IH GG T' Hinvex.
invert Hinvex; [].
intros G' Hinv <-.
so (IH _ _ Hinv) as (<- & <-).
auto.
}
Qed.


Lemma lclose_fun :
  forall P1 P2 P3 P3',
    lclose P1 P2 P3
    -> lclose P1 P2 P3'
    -> P3 = P3'.
Proof.
intros P1 P2 P3 P3' Hclose Hclose'.
revert P3' Hclose'.
induct Hclose.

(* down *)
{
intros T P P3' Hclose.
invert Hclose; [].
auto.
}

(* ex *)
{
intros k P1 P2 P3 _ IH PP Hcloseex.
invert Hcloseex; [].
intros P3' Hclose <-.
so (IH _ Hclose).
subst P3'.
reflexivity.
}
Qed.
    

Lemma prin_fun :
  (forall G e t t',
     print G e t
     -> print G e t'
     -> t = t')
  /\
  (forall G M T T',
     prins G M T
     -> prins G M T'
     -> T = T')
  /\
  (forall G M T T',
     prinps G M T
     -> prinps G M T'
     -> T = T').
Proof.
exploit (prin_ind
           (fun G e t =>
              forall t',
                print G e t' -> t = t')
           (fun G M T =>
              forall T',
                prins G M T' -> T = T')
           (fun G M T =>
              forall T',
                prinps G M T' -> T = T')) as Hind.

(* var *)
intros G i t Hindex t' Hprint.
invert Hprint; [].
intros t'' Hindex' <-.
f_equal; [].
so (index_fun _#5 Hindex Hindex') as H.
injection H.
auto; done.

(* star *)
intros G t' H.
invert H; [].
intros; subst; auto; done.

(* lam *)
intros G t1 t2 e Hoft1 _ IH t' Hprint.
invert Hprint; [].
intros t2' _ Hprint' <-.
f_equal; [].
f_equal; [].
auto; done.

(* app *)
intros G t t1 t2 t1' e1 e2 _ IH1 Hnorm _ _ _ u2 Hprint.
invert Hprint; [].
intros u u1 _ Hprint' Hnorm' _ _.
so (IH1 _ Hprint'); subst u.
so (headnorm_fun _#4 Hnorm Hnorm') as Heq.
injection Heq; auto; done.

(* pair *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 u Hprint.
invert Hprint; [].
intros u1 u2 Hprint1 Hprint2 <-.
so (IH1 _ Hprint1).
so (IH2 _ Hprint2).
f_equal; auto; done.

(* pi1 *)
intros G t t1 t2 e _ IH Hnorm u1 Hprint.
invert Hprint; [].
intros u u2 Hprint' Hnorm'.
so (IH _ Hprint'); subst u.
so (headnorm_fun _#4 Hnorm Hnorm') as Heq.
injection Heq; auto; done.

(* pi2 *)
intros G t t1 t2 e _ IH Hnorm u2 Hprint.
invert Hprint; [].
intros u u1 Hprint' Hnorm'.
so (IH _ Hprint'); subst u.
so (headnorm_fun _#4 Hnorm Hnorm') as Heq.
injection Heq; auto; done.

(* plam *)
intros G k t e _ _ IH _ u Hprint.
invert Hprint; [].
intros t' _ Hprint' _ <-.
f_equal; auto; done.

(* papp *)
intros G u k t e c _ IH Hnorm _ v Hprint.
invert Hprint; [].
intros u' k' t' Hprint' Hnorm' _ <-.
f_equal; [].
so (IH _ Hprint'); subst u'.
so (headnorm_fun _#4 Hnorm Hnorm') as Heq.
injection Heq; auto; done.

(* pack *)
intros G k t c e _ _ _ _ _ _ u Hprint.
invert Hprint; auto; done.

(* unpack *)
intros G u k t1 t2 e1 e2 _ IH1 Hnorm _ IH2 _ v Hprint.
invert Hprint; [].
intros u' k' t1' t2' Hprint1 Hnorm' Hprint2 _ <-.
so (IH1 _ Hprint1); subst u'.
so (headnorm_fun _#4 Hnorm Hnorm') as Heq.
injection Heq.
intros <- <-.
so (IH2 _ Hprint2); subst t2'.
reflexivity.

(* fix *)
intros G t e _ _ _ _ u Hprint.
invert Hprint; [].
auto; done.

(* lett *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 u Hprint.
invert Hprint; [].
intros t1' t2' Hprint1 Hprint2 <-.
so (IH1 _ Hprint1); subst t1'.
so (IH2 _ Hprint2); subst t2'.
reflexivity.

(* letm *)
{
intros G P G' T k t M e _ IH1 Hinv Hfst _ IH2 _ u Hprint.
invert Hprint; [].
intros P' G'' T' k' t' Hprin1 Hinv' Hfst' Hprin2 _ <-.
so (IH1 _ Hprin1); subst P'.
so (linv_fun _#5 Hinv Hinv') as (<- & <-).
so (fsts_fun _#3 Hfst Hfst'); subst k'.
so (IH2 _ Hprin2); subst t'.
reflexivity.
}

(* ext *)
intros G t M _ IH u Hprint.
invert Hprint; [].
intros Hprint'.
so (IH _ Hprint') as Heq.
injection Heq; auto; done.

(* var *)
intros G i T c Hindex U Hprint.
invert Hprint; [].
intros T' c' Hindex' <-.
so (index_fun _#5 Hindex Hindex') as Heq.
injection Heq.
intros <- <-.
reflexivity.

(* star *)
intros G T' Hprins.
invert Hprins; [].
intros <-.
auto; done.

(* stat *)
intros G c k Hprink T' Hprins.
invert Hprins; [].
intros k' Hprink' <-.
so (prink_fun _#4 Hprink Hprink'); subst k'.
reflexivity.

(* dyn *)
intros G e t _ IH T' Hprins.
invert Hprins; [].
intros t' Hprint <-.
so (IH _ Hprint); subst t'.
auto; done.

(* lam *)
intros G T1 T2 k1 M _ Hfst _ IH T' Hprins.
invert Hprins; [].
intros T2' k1' _ Hfst' Hprins' <-.
so (fsts_fun _#3 Hfst Hfst'); subst k1'.
so (IH _ Hprins') as Heq.
f_apply (subst1 nonsense) in Heq as Heq'.
simp_sub in Heq'.
subst T2'.
auto; done.

(* app *)
intros G T1 T2 T1' M1 M2 c2 _ IH _ _ _ Hfstm U Hprins.
invert Hprins; [].
intros U1 U2 U1' c2' Hprins' _ _ Hfstm' <-.
so (IH _ Hprins') as Heq.
injection Heq.
intros <- <-.
so (fstm_fun _#4 Hfstm Hfstm'); subst c2'.
auto; done.

(* pair *)
intros G T1 T2 M1 M2 _ IH1 _ IH2 T' Hprins.
invert Hprins; [].
intros T1' T2' Hprins1 Hprins2 <-.
so (IH1 _ Hprins1) as <-.
so (IH2 _ Hprins2) as <-.
auto; done.

(* pi1 *)
intros G T1 T2 M _ IH T' Hprins.
invert Hprins; [].
intros T2' Hprins'.
so (IH _ Hprins') as Heq.
injection Heq.
intros <- <-.
auto; done.

(* pi2 *)
intros G T1 T2 M c _ IH Hfstm T' Hprins.
invert Hprins; [].
intros T1' T2' c' Hprins' Hfstm' <-.
so (IH _ Hprins') as Heq.
injection Heq.
intros <-.
so (fstm_fun _#4 Hfstm Hfstm'); subst c'.
auto; done.

(* lett *)
intros G t T e M _ IH1 _ IH2 U Hprins.
invert Hprins; [].
intros t' T' Hprin1 Hprin2 <-.
so (IH1 _ Hprin1); subst t'.
so (IH2 _ Hprin2) as <-.
auto; done.

(* circ *)
{
intros G M P _ IH U Hprins.
invert Hprins; [].
intros P' Hprin <-.
so (IH _ Hprin).
subst P'.
reflexivity.
}

(* ret *)
{
intros G M T _ IH U Hprinps.
invert Hprinps; [].
intros T' Hprin <-.
so (IH _ Hprin).
subst T'.
reflexivity.
}

(* seal *)
{
intros G M T T' _ IH Hsubs U Hprinps.
invert Hprinps; [].
auto.
}

(* bind *)
{
intros G M1 M2 P1 G' T k P2 P3 _ IH1 Hinv Hfst _ IH2 Hclose U Hprinps.
invert Hprinps; [].
intros P1' G'' T' k' P2' Hprin1 Hinv' Hfst' Hprin2 Hclose'.
so (IH1 _ Hprin1) as Heq.
injection Heq.
intros <-.
so (linv_fun _#5 Hinv Hinv') as (<- & <-).
so (fsts_fun _#3 Hfst Hfst').
subst k'.
so (IH2 _ Hprin2).
subst P2'.
exact (lclose_fun _#4 Hclose Hclose').
}

(* unpack *)
{
intros G u k t P e M _ IH1 Hnorm _ IH2 U Hprinps.
invert Hprinps; [].
intros u' k' t' P' Hprin1 Hnorm' Hprin2 <-.
so (IH1 _ Hprin1).
subst u'.
so (headnorm_fun _#4 Hnorm Hnorm') as Heq.
injection Heq.
intros <- <-.
so (IH2 _ Hprin2).
subst P'.
reflexivity.
}

(* epilogue *)
{
destruct_all Hind.
do2 2 split; eauto; done.
}
Qed.


Lemma print_fun :
  forall G e t t',
    print G e t
    -> print G e t'
    -> t = t'.
Proof.
exact (prin_fun andel).
Qed.


Lemma prins_fun :
  forall G M T T',
    prins G M T
    -> prins G M T'
    -> T = T'.
Proof.
exact (prin_fun anderl).
Qed.


Lemma prinps_fun :
  forall G M T T',
    prinps G M T
    -> prinps G M T'
    -> T = T'.
Proof.
exact (prin_fun anderr).
Qed.


Inductive oftp : context -> (* tm *) term -> (* cn *) term -> Prop :=
| oftp_var {G i t}
    : index i G (cl_tm t)
      -> oftp G (var i) (subst (sh (S i)) t)

| oftp_star {G}
    : oftp G tm_star cn_unit

| oftp_lam {G t1 t2 e}
    : ofc G t1 kd_type
      -> oftp (G; cl_tm t1) e (shift1 t2)
      -> oftp G (tm_lam t1 e) (cn_arrow t1 t2)

| oftp_app {G t1 t2 e1 e2}
    : oftp G e1 (cn_arrow t1 t2)
      -> oftp G e2 t1
      -> oftp G (tm_app e1 e2) t2

| oftp_pair {G t1 t2 e1 e2}
    : oftp G e1 t1
      -> oftp G e2 t2
      -> oftp G (tm_pair e1 e2) (cn_prod t1 t2)

| oftp_pi1 {G t1 t2 e}
    : oftp G e (cn_prod t1 t2)
      -> oftp G (tm_pi1 e) t1

| oftp_pi2 {G t1 t2 e}
    : oftp G e (cn_prod t1 t2)
      -> oftp G (tm_pi2 e) t2

| oftp_plam {G k t e}
    : ofk G k
      -> oftp (G; cl_cn k) e t
      -> svalue e
      -> oftp G (tm_plam k e) (cn_all k t)

| oftp_papp {G k t e c}
    : oftp G e (cn_all k t)
      -> ofc G c k
      -> oftp G (tm_papp e c) (subst1 c t)

| oftp_pack {G k t c e}
    : ofc G c k
      -> oftp G e (subst1 c t)
      -> ofc (G; cl_cn k) t kd_type
      -> oftp G (tm_pack c e k t) (cn_exists k t)

| oftp_unpack {G k t t' e1 e2}
    : oftp G e1 (cn_exists k t)
      -> oftp (G; cl_cn k; cl_tm t) e2 (subst (sh 2) t')
      -> ofc G t' kd_type
      -> oftp G (tm_unpack e1 e2) t'

| oftp_fix {G t e}
    : oftp G e (cn_arrow (cn_arrow cn_unit t) t)
      -> oftp G (tm_fix t e) t

| oftp_lett {G t1 t2 e1 e2}
    : oftp G e1 t1
      -> oftp (G; cl_tm t1) e2 (shift1 t2)
      -> oftp G (tm_lett e1 e2) t2

| oftp_equiv {G t t' e}
    : oftp G e t
      -> eqc G t t' kd_type
      -> oftp G e t'

| oftp_letm {G M e P G' T k t}
    : oflmp G M P
      -> linv P G' T
      -> fsts T k
      -> oftp (G;; G'; cl_cn k; cl_md (shift1 T) (var 0)) (subst (under 2 (sh (length G'))) e) (subst (sh (2 + length G')) t)
      -> oftp G (tm_letm M e) t

| oftp_ext {G t M}
    : ofmp G M (sg_dyn t)
      -> oftp G (tm_ext M) t


with ofmp : context -> (* md *) term -> (* sg *) term -> Prop :=
| ofmp_var {G i T c}
    : index i G (cl_md T c)
      -> ofmp G (var i) (singles (subst (sh (S i)) c) (subst (sh (S i)) T))

| ofmp_star {G}
    : ofmp G md_star sg_one

| ofmp_stat {G c k}
    : ofc G c k
      -> ofmp G (md_stat c) (sg_stat (single c k))

| ofmp_dyn {G e t}
    : oftp G e t
      -> ofmp G (md_dyn e) (sg_dyn t)

| ofmp_lam {G T1 T2 k1 M}
    : ofs G T1
      -> fsts T1 k1
      -> ofmp (G; cl_cn k1; cl_md (shift1 T1) (var 0)) M (shift1 T2)
      -> ofmp G (md_lam T1 M) (sg_pi T1 T2)

| ofmp_app {G T1 T2 M1 M2 c2}
    : ofmp G M1 (sg_pi T1 T2)
      -> ofmp G M2 T1
      -> fstm G M2 c2
      -> ofmp G (md_app M1 M2) (subst1 c2 T2)

| ofmp_pair {G T1 T2 M1 M2}
    : ofmp G M1 T1
      -> ofmp G M2 T2
      -> ofmp G (md_pair M1 M2) (sg_sigma T1 (shift1 T2))

| ofmp_pi1 {G T1 T2 M}
    : ofmp G M (sg_sigma T1 T2)
      -> ofmp G (md_pi1 M) T1

| ofmp_pi2 {G T1 T2 M c}
    : ofmp G M (sg_sigma T1 T2)
      -> fstm G M c
      -> ofmp G (md_pi2 M) (subst1 (cn_pi1 c) T2)

| ofmp_lett {G t T e M}
    : oftp G e t
      -> ofmp (G; cl_tm t) M (shift1 T)
      -> ofmp G (md_lett e M) T

| ofmp_cir {G M P}
    : oflmp G M P
      -> ofmp G (md_cir M) (sg_circ P)

| ofmp_subsume {G T T' M}
    : ofmp G M T
      -> subs G T T'
      -> ofmp G M T'


with oflmp : context -> (* lm *) term -> (* ps *) term -> Prop :=
| oflmp_ret {G M T}
    : ofmp G M T
      -> oflmp G (lm_ret M) (ps_down T)

| oflmp_seal {G M T}
    : ofmp G M T
      -> oflmp G (lm_seal M T) (ps_down T)

| oflmp_bind {G M1 M2 P1 G' T k P2}
    : ofmp G M1 (sg_circ P1)
      -> linv P1 G' T
      -> fsts T k
      -> oflmp (G;; G'; cl_cn k; cl_md (shift1 T) (var 0)) (subst (under 2 (sh (length G'))) M2) (subst (sh (2 + length G')) P2)
      -> oflmp G (lm_bind M1 M2) P2

| oflmp_unpack {G k t P e M}
    : oftp G e (cn_exists k t)
      -> oflmp (G; cl_cn k; cl_tm t) M (subst (sh 2) P)
      -> oflmp G (lm_unpack e M) P

| oflmp_subsume {G P P' M}
  : oflmp G M P
    -> subps G P P'
    -> oflmp G M P'

.



Scheme oftp_mut_ind := Minimality for oftp Sort Prop
with   ofmp_mut_ind := Minimality for ofmp Sort Prop
with   oflmp_mut_ind := Minimality for oflmp Sort Prop.
Combined Scheme dynamic_principal_ind from oftp_mut_ind, ofmp_mut_ind, oflmp_mut_ind.


Lemma ofxp_ofx :
  (forall G e t,
     wf G
     -> oftp G e t
     -> oft G e t)
  /\
  (forall G M T,
     wf G
     -> ofmp G M T
     -> ofm G M T)
  /\
  (forall G M T,
     wf G
     -> oflmp G M T
     -> oflm G M T).
Proof.
exploit
  (dynamic_principal_ind
     (fun G e t => wf G -> oft G e t)
     (fun G M T => wf G -> ofm G M T)
     (fun G M T => wf G -> oflm G M T)) as Hind;
try (intros; eauto with static; done).

(* unpack *)
intros G k t t' e1 e2 _ IH1 _ IH2 Hoft' Hwf.
so (IH1 Hwf) as Hof1.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf Hof1)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
eauto with static; done.

(* lett *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 Hwf.
so (oft_reg _#3 Hwf (IH1 Hwf)).
eauto with static; done.


(* letm *)
{
intros G M e P G' T k t _ IH1 Hinv Hfst _ IH2 Hwf.
so (IH1 Hwf) as HofM.
so (oflm_reg _#3 Hwf HofM) as HofP.
so (linv_reg _#4 Hwf HofP Hinv) as (Hwf' & HofT).
so (IH2 (wf_md Hwf' HofT Hfst)) as Hofe.
simp_sub in Hofe.
eapply oft_letm; eauto.
}

(* var *)
intros G i T c Hindex Hwf.
eapply singles_ofm; eauto with static sequence; done.

(* stat *)
intros G c k Hof Hwf.
apply ofm_stat; [].
eapply single_ofc; eauto; done.

(* lett *)
intros G t T e M _ IH1 _ IH2 Hwf.
so (oft_reg _#3 Hwf (IH1 Hwf)).
eauto with static; done.

(* bind *)
{
intros G M1 M2 P1 G' T k P2 _ IH1 Hinv Hfst _ IH2 Hwf.
so (IH1 Hwf) as HofM1.
so (ofm_reg _#3 Hwf HofM1) as Hofcirc.
invert Hofcirc.
intro HofP1.
so (linv_reg _#4 Hwf HofP1 Hinv) as (Hwf' & HofT).
lapply IH2; auto using wf_md.
intros HofM2.
eapply oflm_bind; eauto.
}

(* unpack *)
{
intros G k t P e M _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as Hofe.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf Hofe)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
eauto with static; done.
}

(* epilogue *)
{
destruct_all Hind.
do2 2 split; auto; done.
}
Qed.


Lemma oftp_oft :
  forall G e t,
    wf G
    -> oftp G e t
    -> oft G e t.
Proof.
exact (ofxp_ofx andel).
Qed.


Lemma ofmp_ofm :
  forall G M T,
    wf G
    -> ofmp G M T
    -> ofm G M T.
Proof.
exact (ofxp_ofx anderl).
Qed.


Lemma oflmp_oflm :
  forall G M T,
    wf G
    -> oflmp G M T
    -> oflm G M T.
Proof.
exact (ofxp_ofx anderr).
Qed.


Lemma lclose_reg :
  forall G P1 P2 P3 G' T,
    linv P1 G' T
    -> ofps G P1
    -> ofps (G;; G') P2
    -> lclose P1 P2 P3
    -> ofps G P3.
Proof.
intros G P1 P2 P3 G' T Hinv HofP1 HofP2 Hclose.
revert G P2 P3 HofP1 HofP2 Hclose.
induct Hinv.

(* down *)
{
intros T G P2 P3 HofP1 HofP2 Hclose.
invert Hclose; [].
intros <-.
simpl in HofP2.
exact HofP2.
}

(* ex *)
{
intros k P G' T _ IH G P2 PP Hofex HofP2 Hclose.
invert Hclose; [].
intros P3 Hclose' <-.
rewrite <- app_assoc in HofP2.
simpl app in HofP2.
invert Hofex; [].
intros Hofk HofP1.
so (IH _#3 HofP1 HofP2 Hclose') as HofP3.
apply ofps_ex; auto.
}
Qed.


Lemma lclose_subps :
  forall G P1 P2 P3 G' T,
    linv P1 G' T
    -> ofps G P1
    -> ofps (G;; G') P2
    -> lclose P1 P2 P3
    -> subps (G;; G') P2 (subst (sh (length G')) P3).
Proof.
intros G P1 P2 P3 G' T Hinv HofP1 HofP2 Hclose.
revert G P2 P3 HofP1 HofP2 Hclose.
induct Hinv.

(* down *)
{
intros T G P2 P3 HofP1 HofP2 Hclose.
invert Hclose; [].
intros <-.
simpl.
simp_sub.
apply subps_refl.
apply eqps_refl; auto.
}

(* ex *)
{
intros k P1 G' T Hinv IH G P2 PP Hofex HofP2 Hclose.
invert Hofex; [].
intros Hofk HofP1.
invert Hclose; [].
intros P3 Hclose' <-.
rewrite <- app_assoc.
simpl app.
rewrite -> app_length.
simpl length.
simp_sub.
rewrite <- app_assoc in HofP2.
simpl length in HofP2.
so (IH _#3 HofP1 HofP2 Hclose') as Hsubs.
apply (@subps_ex_right _#4 (var (length G'))).
  {
  apply ofc_var.
  replace (length G') with (0 + length G') by omega.
  apply index_app_right.
  apply index_0.
  }

  {
  replace (2 + length G') with (S (S (length G'))) by omega.
  simp_sub.
  rewrite <- (subst_dot0_sh P3) in Hsubs.
  rewrite <- subst_compose in Hsubs.
  rewrite -> compose_dot in Hsubs.
  simp_sub in Hsubs.
  exact Hsubs.
  }

  {
  simpl app in HofP2.
  so (lclose_reg _#6 Hinv HofP1 HofP2 Hclose') as HofP3.
  eapply subst_ofps; eauto.
  apply subof_dot.
    {
    apply subof_sh.
    apply truncate_S.
    replace (S (length G')) with (length G' + 1) by omega.
    eapply truncate_sum.
      {
      apply app_truncate.
      }
    apply truncate_S.
    apply truncate_0.
    }
  simp_sub.
  apply sof_cn.
  eapply rel_quest.
    {
    eapply ofc_var.
    apply index_0.
    }
  simp_sub.
  reflexivity.
  }
}
Qed.


Lemma prinx_ofxp :
  (forall G e t,
     wf G
     -> print G e t
     -> oftp G e t)
  /\
  (forall G M T,
     wf G
     -> prins G M T
     -> ofmp G M T)
  /\
  (forall G M T,
     wf G
     -> prinps G M T
     -> oflmp G M T).
Proof.
exploit (prin_ind
           (fun G e t => wf G -> oftp G e t)
           (fun G M T => wf G -> ofmp G M T)
           (fun G M T => wf G -> oflmp G M T)) as Hind;
try (intros; eauto using oftp, ofmp, oflmp, ofc_reg; done).

(* lam *)
intros G t1 t2 e Hofc _ IH Hwf.
so (wf_tm Hwf Hofc) as Hwf'.
so (IH Hwf') as Hofe.
so (oft_reg _#3 Hwf' (oftp_oft _#3 Hwf' Hofe)) as Hoft2.
apply oftp_lam; auto; [].
simp_sub.
erewrite <- subord_tm_cn; eauto; done.

(* app *)
intros G t t1 t2 t1' e1 e2 _ IH1 Hnorm _ IH2 Heqc2 Hwf.
so (IH1 Hwf) as Hofe1.
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe1)) as Hoft.
so (headnorm_sound _#4 Hwf Hoft Hnorm) as Heqc1.
eapply oftp_app; eauto using oftp_equiv; done.

(* pi1 *)
intros G t t1 t2 e _ IH Hnorm Hwf.
so (IH Hwf) as Hofe.
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe)) as Hoft.
so (headnorm_sound _#4 Hwf Hoft Hnorm) as Heqc.
eapply oftp_pi1; eauto using oftp_equiv; done.

(* pi2 *)
intros G t t1 t2 e _ IH Hnorm Hwf.
so (IH Hwf) as Hofe.
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe)) as Hoft.
so (headnorm_sound _#4 Hwf Hoft Hnorm) as Heqc.
eapply oftp_pi2; eauto using oftp_equiv; done.

(* plam *)
intros G k t e Hofk _ IH Hval Hwf.
so (IH (wf_cn Hwf Hofk)) as Hofe.
apply oftp_plam; auto; done.

(* papp *)
intros G u k t e c _ IH1 Hnorm Hofc Hwf.
so (IH1 Hwf) as Hofe.
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe)) as Hofu.
so (headnorm_sound _#4 Hwf Hofu Hnorm) as Heqc.
eapply oftp_papp; eauto using oftp_equiv; done.

(* unpack *)
intros G u k t t' e1 e2 _ IH1 Hnorm _ IH2 Heqinh Hwf.
so (IH1 Hwf) as Hofe1.
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe1)) as Hofu.
so (headnorm_sound _#4 Hwf Hofu Hnorm) as Hequ.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Hequ ander andel)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _; clear l Hofcr.
so (wf_tm (wf_cn Hwf Hofk) Hoft) as Hwf'.
so (IH2 Hwf') as Hofe2.
so (oft_reg _#3 Hwf' (oftp_oft _#3 Hwf' Hofe2)) as Hoft'.
#2 eapply oftp_unpack; eauto using oftp_equiv.
  simp_sub.
  eapply oftp_equiv; eauto; [].
  repl t' with (subst (dot nonsense sh1) t') at 1.
    symmetry.
    eapply subord_tm_cn; eauto; done.
  so (subst_eqc _#6 (weaken1 _ _ (cl_tm t)) Heqinh) as Heqinh'.
  simp_sub in Heqinh'.
  assumption.

  eapply subst_ofc_type; eauto; [].
  #2 apply subof_dot.
    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply sof_cn; [].
    apply inh_ofc; auto; done.

    simp_sub.
    apply sof_tm; done.

(* lett *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as Hofe1.
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe1)) as Hoft1.
so (wf_tm Hwf Hoft1) as Hwf'.
so (IH2 Hwf') as Hofe2.
so (oft_reg _#3 Hwf' (oftp_oft _#3 Hwf' Hofe2)) as Hoft2.
eapply oftp_lett; eauto; [].
simp_sub.
erewrite <- subord_tm_cn; eauto; done.

(* letm *)
{
intros G P G' T k t M e _ IH1 Hinv Hfst _ IH2 Heqc Hwf.
so (IH1 Hwf) as HofM.
so (oflm_reg _#3 Hwf (oflmp_oflm _#3 Hwf HofM)) as HofP.
so (linv_reg _#4 Hwf HofP Hinv) as (Hwf' & HofT).
so (wf_md Hwf' HofT Hfst) as Hwf''.
so (IH2 Hwf'') as Hofe.
so (oft_reg _#3 Hwf'' (oftp_oft _#3 Hwf'' Hofe)) as Hoft.
eapply oftp_letm; eauto.
eapply oftp_equiv; eauto.
repl t with (subst sh1 (subst (dot nonsense id) t)) at 1.
  {
  simp_sub.
  symmetry.
  eapply subord_md_cn; eauto.
  }
simp_sub.
so (subst_eqc_type _#5 (weaken1 _ _ (cl_md (shift1 T) (var 0))) Heqc) as Heqc'.
simp_sub in Heqc'.
exact Heqc'.
}

(* stat *)
intros G c k Hprink Hwf.
#2 apply (@ofmp_subsume _ (sg_stat (single c k))).
  apply ofmp_stat; auto using prink_sound; done.

  apply subs_stat; [].
  apply single_subk; auto using prink_sound; done.

(* lam *)
intros G T1 T2 k1 M HofT1 Hfst _ IH Hwf.
so (IH (wf_md Hwf HofT1 Hfst)) as HofM.
eapply ofmp_lam; eauto; done.

(* lett *)
intros G t T e M _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as Hofe.
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe)) as Hoft.
so (wf_tm Hwf Hoft) as Hwf'.
so (IH2 Hwf') as HofM.
so (ofm_reg _#3 Hwf' (ofmp_ofm _#3 Hwf' HofM)) as HofT.
eapply ofmp_lett; eauto; [].
simp_sub.
erewrite <- subord_tm_sg; eauto; done.

(* bind *)
{
intros G M1 M2 P1 G' T k P2 P2' _ IH1 Hinv Hfst _ IH2 Hclose Hwf.
so (IH1 Hwf) as HofM1.
so (ofm_reg _#3 Hwf (ofmp_ofm _#3 Hwf HofM1)) as Hofcirc.
invert Hofcirc; [].
intro HofP1.
so (linv_reg _#4 Hwf HofP1 Hinv) as (Hwf' & HofT).
assert (wf (G;; G'; cl_cn k; cl_md (shift1 T) (var 0))) as Hwf'' by (auto using wf_md).
so (IH2 Hwf'') as HofM2.
eapply (@oflmp_bind _#7 P2'); eauto.
eapply oflmp_subsume; eauto.
so (oflm_reg _#3 Hwf'' (oflmp_oflm _#3 Hwf'' HofM2)) as HofP2.
so (subord_md_ps _#4 HofP2) as Hsubord.
apply (@subps_trans _ _ (ps_ex (subst (sh 2) k) (subst (dot nonsense (dot (var 0) (sh 3))) P2))).
  {
  apply (@subps_ex_right _#4 (var 1)).
    {
    apply ofc_var; auto with sequence.
    }
  
    {
    simp_sub.
    replace (subst (dot nonsense (dot (var 1) (sh 2))) P2) with P2.
      {
      apply subps_refl.
      apply eqps_refl; auto.
      }
    rewrite -> Hsubord at 1.
    rewrite <- (subst_dot01_sh P2) at 1.
    simp_sub.
    reflexivity.
    }
  
    {
    eapply subst_ofps; eauto.
    apply subof_dot.
      {
      eapply subof_dot; eauto using subof_sh, truncate_S, truncate_0.
      simp_sub.
      apply sof_cn.
      eapply rel_quest.
        {
        eapply ofc_var; eauto using index_0.
        }
      simp_sub.
      reflexivity.
      }
    simp_sub.
    apply sof_md.
    }
  }
eassert Prop as Hsubs; [refine (lclose_subps _#6 Hinv HofP1 _ Hclose) |].
  {
  apply ofps_ex; eauto using ofs_fst.
  eapply subst_ofps; eauto.
  apply subof_dot; auto using subof_id.
  simp_sub.
  apply sof_md.
  }
so (weaken2_subps _ (cl_cn k) (cl_md (shift1 T) (var 0)) _ _ Hsubs) as Hsubs'.
simp_sub in Hsubs'.
exact Hsubs'.
}

(* unpack *)
{
  intros G u k t P e M _ IH1 Hnorm _ IH2 Hwf.
so (IH1 Hwf) as Hofe.
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe)) as Hofu.
so (headnorm_sound _#4 Hwf Hofu Hnorm) as Hequ.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Hequ ander andel)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _; clear l Hofcr.
so (wf_tm (wf_cn Hwf Hofk) Hoft) as Hwf'.
so (IH2 Hwf') as HofM.
eapply oflmp_unpack; eauto using oftp_equiv; [].
simp_sub.
eapply oflmp_subsume; eauto.
so (oflm_reg _#3 Hwf' (oflmp_oflm _#3 Hwf' HofM)) as HofP.
so (subord_tm_ps _#3 HofP) as Hsubord.
apply (@subps_ex_right _#4 (var 1)).
  {
  apply ofc_var; auto with sequence.
  }

  {
  simp_sub.
  replace (subst (dot nonsense (dot (var 1) (sh 2))) P) with P.
    {
    apply subps_refl.
    apply eqps_refl; auto.
    }
  rewrite <- (subst_dot01_sh P) at 1.
  rewrite -> Hsubord at 1.
  simp_sub.
  reflexivity.
  }

  {
  eapply subst_ofps; eauto.
  apply subof_dot.
    {
    apply subof_dot; eauto using subof_sh, truncate_S, truncate_0.
    simp_sub.
    apply sof_cn.
    eapply rel_quest.
      {
      eapply ofc_var; eauto using index_0.
      }
    simp_sub.
    reflexivity.
    }
  simp_sub.
  apply sof_tm.
  }
}

(* epilogue *)
destruct_all Hind.
do2 2 split; auto; done.
Qed.


Lemma print_oftp :
  forall G e t,
    wf G
    -> print G e t
    -> oftp G e t.
Proof.
exact (prinx_ofxp andel).
Qed.


Lemma prins_ofmp :
  forall G M T,
    wf G
    -> prins G M T
    -> ofmp G M T.
Proof.
exact (prinx_ofxp anderl).
Qed.


Lemma prinps_oflmp :
  forall G M T,
    wf G
    -> prinps G M T
    -> oflmp G M T.
Proof.
exact (prinx_ofxp anderr).
Qed.


Lemma print_sound :
  forall G e t,
    wf G
    -> print G e t
    -> oft G e t.
Proof.
intros.
auto using print_oftp, oftp_oft.
Qed.


Lemma prins_sound :
  forall G M T,
    wf G
    -> prins G M T
    -> ofm G M T.
Proof.
intros.
auto using prins_ofmp, ofmp_ofm.
Qed.


Lemma prinps_sound :
  forall G M T,
    wf G
    -> prinps G M T
    -> oflm G M T.
Proof.
intros.
auto using prinps_oflmp, oflmp_oflm.
Qed.


Lemma prins_prink_fst :
  forall G M T c k,
    wf G
    -> fstm G M c
    -> prins G M T
    -> prink G c k
    -> fsts T k.
Proof.
intros G M T c k Hwf HfstM Hprins Hprink.
revert c k Hwf HfstM Hprink.
induct Hprins.

(* var *)
intros G i T c Hindex d l Hwf HfstM Hprink.
so (fstm_fun _#4 HfstM (fstm_var Hindex)); subst d.
assert (exists T' k, T = shift1 T' /\ hsg T' /\ c = var 0 /\ fsts T' k /\ index (S i) G (cl_cn k)) as H.
  clear HfstM Hprink.
  remember (cl_md T c) as C eqn:HeqC.
  revert Hwf HeqC.
  #2 induct Hindex.
    intros C G Hwf ->.
    invert Hwf; [].
    intros G' k T' _ Hof Hfst <- <- <-.
    exists T', k.
    do2 4 split; eauto using ofs_hsg with sequence; done.

    intros i D G C _ IH Hwf HeqC.
    so (IH (wf_tail _ _ Hwf) HeqC) as (T' & k & HeqT & Hhsg & Heqc & Hfsts & Hindex).
    exists T', k.
    do2 4 split; auto.
    apply index_S; auto; done.
destruct H as (T' & k & -> & Hhsg & -> & Hfst & Hindex').
simp_sub.
simp_sub in Hprink.
so (prink_fun _#4 Hprink (prink_var Hindex')); subst l.
apply singles_fst; [].
eapply subst_fsts; eauto; done.

(* star *)
intros G c k Hwf HfstM Hprink.
invert HfstM; [].
intros <-.
so (prink_fun _#4 Hprink prink_star); subst k.
apply fsts_one; done.

(* stat *)
intros G c k Hprink c' k' Hwf HfstM Hprink'.
invert HfstM; [].
intros <-.
so (prink_fun _#4 Hprink Hprink'); subst k'.
apply fsts_stat; done.

(* dyn *)
intros G e t _ c k Hwf HfstM Hprink.
invert HfstM; [].
intros <-.
so (prink_fun _#4 Hprink prink_star); subst k.
apply fsts_dyn; done.

(* lam *)
intros G T1 T2 k1 M Hofs Hfst _ IH d k Hwf Hfstapp Hprink.
invert Hfstapp; [].
intros k1' c Hfst' HfstM <-.
so (fsts_fun _#3 Hfst Hfst'); subst k1'; clear Hfst'.
invert Hprink; [].
intros k2 Hofk1 Hprink' <-.
apply fsts_pi; trivial; [].
#3 sor (refine (IH (shift1 c) (shift1 k2) _ HfstM _)) as Hfst2.
  apply wf_md; auto; done.

  eapply weaken1_prink; eauto; done.

  so (subst_fsts _ _ (dot nonsense id) Hfst2) as Hfst2'.
  simp_sub in Hfst2'.
  assumption.

(* app *)
intros G T1 T2 T1' M1 M2 c2 _ IH1 _ IH2 Hsub HfstM2 c k Hwf Hfstapp Hprink.
invert Hfstapp; [].
intros c1 c2' HfstM1 HfstM2' <-.
so (fstm_fun _#4 HfstM2 HfstM2'); subst c2'; clear HfstM2'.
invert Hprink; [].
intros k1 k2 Hprink1 Hofc2 <-.
so (IH1 _ _ Hwf HfstM1 Hprink1) as Hfstpi.
invert Hfstpi; [].
intros Hfst1 Hfst2.
apply subst_fsts; auto; done.

(* pair *)
intros G T1 T2 M1 M2 _ IH1 _ IH2 c k Hwf HfstM Hprink.  
invert HfstM; [].
intros c1 c2 HfstM1 HfstM2 <-.
invert Hprink; [].
intros k1 k2 Hprink1 Hprink2 <-.
so (IH1 _ _ Hwf HfstM1 Hprink1) as Hfst1.
so (IH2 _ _ Hwf HfstM2 Hprink2) as Hfst2.
apply fsts_sigma; eauto using subst_fsts; done.

(* pi1 *)
intros G T1 T2 M _ IH d k Hwf Hfstpi Hprink.
invert Hfstpi; [].
intros c HfstM <-.
invert Hprink; [].
intros k2 Hprink'.
so (IH _ _ Hwf HfstM Hprink') as Hfsts.
invert Hfsts; auto; done.

(* pi2 *)
intros G T1 T2 M c _ IH HfstM d k Hwf Hfstpi Hprink.
invert Hfstpi; [].
intros c' HfstM' <-.
so (fstm_fun _#4 HfstM HfstM'); subst c'.
invert Hprink; [].
intros k1 k2 Hprink' <-.
so (IH _ _ Hwf HfstM Hprink') as Hfsts.
invert Hfsts; eauto using subst_fsts; done.

(* lett *)
intros G t T e M Hprint _ IH c k Hwf Hfstlet Hprink.
invert Hfstlet; [].
intros HfstM.
so (oft_reg _#3 Hwf (print_sound _#3 Hwf Hprint)) as Hoft.
so (subst_fstm _#5 (subof_tm_utm_tof _ t) HfstM) as HfstM'.
simp_sub in HfstM'.
so (IH _ _ (wf_tm Hwf Hoft) HfstM' (weaken1_prink _#4 Hprink)) as Hfsts.
so (subst_fsts _ _ (dot nonsense id) Hfsts) as Hfsts'.
simp_sub in Hfsts'.
assumption.

(* cir *)
intros G M P H c k Hwf Hfst Hprink.
invert Hfst; [].
intros <-.
invert Hprink; [].
intros <-.
apply fsts_circ.
Qed.


Lemma arrow_headnorm :
  forall G t t1 t2,
    wf G
    -> eqc G t (cn_arrow t1 t2) kd_type
    -> exists t1' t2',
         headnorm G t (cn_arrow t1' t2')
         /\ eqc G t1' t1 kd_type
         /\ eqc G t2' t2 kd_type.
Proof.
intros G t t1 t2 Hwf Heqc.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc ander andel)) as (l & Hofcr).
invert Hofcr; [].
intros Hoft1 Hoft2 _.
so (aeq_complete _#4 Hwf Heqc) as Haeq.
invert Haeq; [].
intros p p' Hnorm Hnorm' Hpeq.
so (headnorm_fun _#4 Hnorm' (headnorm_done natk_arrow)); subst p'.
invert Hpeq; [].
intros t1' t2' Haeq1 Haeq2 <-.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf (eqc_reg _#4 Hwf Heqc andel) Hnorm) ander andel)) as (l' & Hofcr').
invert Hofcr'; [].
intros Hoft1' Hoft2' _.
exists t1', t2'.
do2 2 split.
  assumption.

  eapply aeq_sound; eauto using ctxeq_refl with static; done.

  eapply aeq_sound; eauto using ctxeq_refl with static; done.
Qed.


Lemma prod_headnorm :
  forall G t t1 t2,
    wf G
    -> eqc G t (cn_prod t1 t2) kd_type
    -> exists t1' t2',
         headnorm G t (cn_prod t1' t2')
         /\ eqc G t1' t1 kd_type
         /\ eqc G t2' t2 kd_type.
Proof.
intros G t t1 t2 Hwf Heqc.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc ander andel)) as (l & Hofcr).
invert Hofcr; [].
intros Hoft1 Hoft2 _.
so (aeq_complete _#4 Hwf Heqc) as Haeq.
invert Haeq; [].
intros p p' Hnorm Hnorm' Hpeq.
so (headnorm_fun _#4 Hnorm' (headnorm_done natk_prod)); subst p'.
invert Hpeq; [].
intros t1' t2' Haeq1 Haeq2 <-.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf (eqc_reg _#4 Hwf Heqc andel) Hnorm) ander andel)) as (l' & Hofcr').
invert Hofcr'; [].
intros Hoft1' Hoft2' _.
exists t1', t2'.
do2 2 split.
  assumption.

  eapply aeq_sound; eauto using ctxeq_refl with static; done.

  eapply aeq_sound; eauto using ctxeq_refl with static; done.
Qed.


Lemma all_headnorm :
  forall G u k t,
    wf G
    -> eqc G u (cn_all k t) kd_type
    -> exists k' t',
         headnorm G u (cn_all k' t')
         /\ eqk G k' k
         /\ eqc (G; cl_cn k) t' t kd_type.
Proof.
intros G u k t Hwf Heqc.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc ander andel)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
so (aeq_complete _#4 Hwf Heqc) as Haeq.
invert Haeq; [].
intros p p' Hnorm Hnorm' Hpeq.
so (headnorm_fun _#4 Hnorm' (headnorm_done natk_all)); subst p'.
invert Hpeq; [].
intros k' t' Haeq1 Haeq2 <-.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf (eqc_reg _#4 Hwf Heqc andel) Hnorm) ander andel)) as (l' & Hofcr').
invert Hofcr'; [].
intros Hofk' Hoft' _.
exists k', t'.
do2 2 split.
  assumption.

  eapply aeqk_sound; eauto using ctxeq_refl with static; done.

  apply eqc_symm; [].
  eapply aeq_sound; eauto using aeq_symm with static; [].
  apply ctxeq_cn; auto using ctxeq_refl; [].
  apply eqk_symm; [].
  eapply aeqk_sound; eauto using ctxeq_refl with static; done.
Qed.


Lemma exists_headnorm :
  forall G u k t,
    wf G
    -> eqc G u (cn_exists k t) kd_type
    -> exists k' t',
         headnorm G u (cn_exists k' t')
         /\ eqk G k' k
         /\ eqc (G; cl_cn k) t' t kd_type.
Proof.
intros G u k t Hwf Heqc.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc ander andel)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
so (aeq_complete _#4 Hwf Heqc) as Haeq.
invert Haeq; [].
intros p p' Hnorm Hnorm' Hpeq.
so (headnorm_fun _#4 Hnorm' (headnorm_done natk_exists)); subst p'.
invert Hpeq; [].
intros k' t' Haeq1 Haeq2 <-.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf (eqc_reg _#4 Hwf Heqc andel) Hnorm) ander andel)) as (l' & Hofcr').
invert Hofcr'; [].
intros Hofk' Hoft' _.
exists k', t'.
do2 2 split.
  assumption.

  eapply aeqk_sound; eauto using ctxeq_refl with static; done.

  apply eqc_symm; [].
  eapply aeq_sound; eauto using aeq_symm with static; [].
  apply ctxeq_cn; auto using ctxeq_refl; [].
  apply eqk_symm; [].
  eapply aeqk_sound; eauto using ctxeq_refl with static; done.
Qed.


(* no dependent sigmas *)
Inductive ndss : (* sg *) term -> Prop :=
| ndss_one
    : ndss sg_one

| ndss_stat {k}
    : ndss (sg_stat k)

| ndss_dyn {t}
    : ndss (sg_dyn t)

| ndss_pi {T1 T2}
    : ndss T2
      -> ndss (sg_pi T1 T2)

| ndss_sigma {T1 T2}
    : ndss T1
      -> ndss T2
      -> ndss (sg_sigma T1 (shift1 T2))

| ndss_circ {P}
    : ndss (sg_circ P)
.


Lemma subst_ndss :
  forall s T,
    ndss T
    -> ndss (subst s T).
Proof.
intros s T Hnds.
revert s.
induct Hnds; try (intros; simp_sub; eauto using ndss; done); [].

(* sigma *)
intros T1 T2 _ IH1 _ IH2 s.
simp_sub.
rewrite -> subst_compose; [].
apply ndss_sigma; auto; done.
Qed.


Lemma single_ndss :
  forall c T, ndss (singles c T).
Proof.
intros c T.
revert c.
induct T; try (intros; simpl; eauto using ndss; done); [].
(* sigma *)
intros T1 IH1 T2 IH2 c.
simpl.
replsub (dot (cn_pi1 (shift1 c)) sh1) with (compose (dot (cn_pi1 c) id) sh1).
rewrite -> subst_compose; [].
apply ndss_sigma; auto using subst_ndss; done.
Qed.


Lemma prins_ndss :
  forall G M T,
    prins G M T
    -> ndss T.
Proof.
intros G M T H.
induct H;
try (intros; eauto using ndss; discriminate).

(* var *)
intros.
apply single_ndss; done.

(* lamap *)
intros G T1 T2 k1 M _ _ _ IH.
so (subst_ndss (dot nonsense id) _ IH) as H'.
simp_sub in H'.
apply ndss_pi; auto; done.

(* appap *)
intros G T1 T2 T1' M1 M2 c2 _ IH _ _ _ _.
invert IH; [].
intros H'.
eapply subst_ndss; eauto; done.

(* pi1 *)
intros G T1 T2 M _ IH.
invert IH; [].
auto; done.

(* pi2 *)
intros G T1 T2 M c _ IH _.
invert IH; [].
intros T2' _ H' <-.
simp_sub.
assumption.

(* lett *)
intros G t T e M _ _ IH.
eapply subst_ndss; eauto; done.
Qed.


Inductive ctxsub : context -> sub -> context -> Prop :=
| ctxsub_nil {G}
    : ctxsub G (sh (length G)) nil

| ctxsub_cn {G c s G' k}
    : ctxsub G s G'
      -> ofc G c (subst s k)
      -> ctxsub G (dot c s) (G'; cl_cn k)

| ctxsub_tm {G i s G' t t'}
    : ctxsub G s G'
      -> index i G (cl_tm t)
      -> eqc G (subst (sh (S i)) t) (subst s t') kd_type
      -> ctxsub G (dot (var i) s) (G'; cl_tm t')
    
| ctxsub_md {G i s G' T T' c c'}
    : ctxsub G s G'
      -> index i G (cl_md T c)
      -> subs G (subst (sh (S i)) T) (subst s T')
      -> subst (sh (S i)) c = subst s c'
      -> ctxsub G (dot (var i) s) (G'; cl_md T' c')
.


Lemma ctxsub_subof :
  forall G s G',
    ctxsub G s G'
    -> subof dof G s G'.
Proof.
intros G s G' Hs.
induct Hs.

(* nil *)
{
intros G.
apply subof_sh.
apply truncate_all.
}

(* cn *)
{
intros G c s G' k _ IH Hofc.
apply subof_dot; auto.
simp_sub.
apply dof_cn; auto.
}

(* tm *)
{
intros G i s G' t t' _ IH Hindex Heqt.
apply subof_dot; auto.
simp_sub.
apply dof_tm.
eapply oft_equiv; eauto.
apply oft_var; auto.
}

(* md *)
{
intros G i s G' T T' c c' _ IH Hindex Hsubs Heqc.
apply subof_dot; auto.
simp_sub.
apply dof_md.
  {
  eapply ofm_subsume; eauto.
  eapply ofm_var; eauto.
  }

  {
  rewrite <- Heqc.
  eapply fstm_var; eauto.
  }
}
Qed.


Lemma ctxsub_tail :
  forall G s G' A,
    ctxsub G s (G'; A)
    -> exists m s',
         ctxsub G s' G'
         /\ s = dot m s'.
Proof.
intros G s G' A Hs.
invert Hs; intros; subst; eauto.
Qed.


Lemma ctxsub_lookup_tm :
  forall G s G' i t',
    ctxsub G s G'
    -> index i G' (cl_tm t')
    -> exists j t,
         project s i = (var j)
         /\ index j G (cl_tm t)
         /\ eqc G (subst (sh (S j)) t) (subst (trunc (S i) s) t') kd_type.
Proof.
intros G s G' i t' Hs Hindex.
revert G s G' Hs Hindex.
induct i.

(* 0 *)
{
intros G s GG Hs Hindex.
invert Hindex; [].
intros G' <-.
invert Hs; [].
intros j s' t Hs' Hindex' Heqt <-.
exists j, t.
do2 2 split; auto.
}

(* S *)
{
intros i IH G s GG Hs Hindex.
invert Hindex; [].
intros A G' Hindex' <-.
so (ctxsub_tail _#4 Hs) as (m & s' & Hs' & ->).
so (IH _ _ _ Hs' Hindex') as (j & t & Hproject & Hindex'' & Heqt).
exists j, t.
do2 2 split; auto.
}
Qed.


Lemma ctxsub_lookup_md :
  forall G s G' i T' c',
    ctxsub G s G'
    -> index i G' (cl_md T' c')
    -> exists j T c,
         project s i = (var j)
         /\ index j G (cl_md T c)
         /\ subs G (subst (sh (S j)) T) (subst (trunc (S i) s) T')
         /\ subst (sh (S j)) c = subst (trunc (S i) s) c'.
Proof.
intros G s G' i T' c' Hs Hindex.
revert G s G' Hs Hindex.
induct i.

(* 0 *)
{
intros G s GG Hs Hindex.
invert Hindex; [].
intros G' <-.
invert Hs; [].
intros j s' T c Hs' Hindex' HsubT Heqc <-.
exists j, T, c.
do2 3 split; auto.
}

(* S *)
{
intros i IH G s GG Hs Hindex.
invert Hindex; [].
intros A G' Hindex' <-.
so (ctxsub_tail _#4 Hs) as (m & s' & Hs' & ->).
so (IH _ _ _ Hs' Hindex') as (j & T & c & Hproject & Hindex'' & HsubT & Heqc).
exists j, T, c.
do2 3 split; auto.
}
Qed.


Lemma ctxsub_weaken :
  forall G1 s G2 G,
    ctxsub G1 s G2
    -> ctxsub (G1;; G) (compose s (sh (length G))) G2.
Proof.
intros G1 s G2 G Hs.
induct Hs.

(* nil *)
{
intros G1.
simp_sub.
replace (length G1 + length G) with (length G + length G1) by omega.
rewrite <- app_length.
apply ctxsub_nil.
}

(* cn *)
{
intros G1 c s G2 k _ IH Hofc.
simp_sub.
apply ctxsub_cn; auto.
rewrite -> subst_compose.
eapply subst_ofc; eauto.
apply subof_sh.
apply app_truncate.
}

(* tm *)
{
intros G1 i s G2 t t' _ IH Hindex Heqt.
simp_sub.
eapply ctxsub_tm; eauto.
  {
  replace (length G + i) with (i + length G) by omega.
  apply index_app_right; eauto.
  }
replace (S (length G + i)) with (S i + length G) by omega.
rewrite <- compose_sh_sh.
rewrite -> !subst_compose.
eapply subst_eqc_type; eauto.
apply subof_sh.
apply app_truncate.
}

(* md *)
{
intros G1 i s G2 T T' c c' _ IH Hindex HsubT Heqc.
simp_sub.
eapply ctxsub_md; eauto.
  {
  replace (length G + i) with (i + length G) by omega.
  apply index_app_right; eauto.
  }

  {
  replace (S (length G + i)) with (S i + length G) by omega.
  rewrite <- compose_sh_sh.
  rewrite -> !subst_compose.
  eapply subst_subs; eauto.
  apply subof_sh.
  apply app_truncate.
  }

  {
  replace (S (length G + i)) with (S i + length G) by omega.
  rewrite <- compose_sh_sh.
  rewrite -> !subst_compose.
  f_equal; auto.
  }
}
Qed.


Lemma ctxsub_all :
  forall G s G',
    ctxsub G s G'
    -> compose (sh (length G')) s = sh (length G).
Proof.
intros G s G' H.
induct H; intros; simpl length; simp_sub; auto.
Qed.


Lemma ctxsub_trans :
  forall G1 G2 G3 s s',
    ctxsub G1 s G2
    -> ctxsub G2 s' G3
    -> ctxsub G1 (compose s' s) G3.
Proof.
intros G1 G2 G3 s s' H12 H23.
revert H12.
induct H23.

(* nil *)
{
intros G2 Hs.
erewrite -> ctxsub_all; eauto.
apply ctxsub_nil.
}

(* cn *)
{
intros G2 c s' G3 k _ IH Hofc H12.
simp_sub.
apply ctxsub_cn; auto.
rewrite -> subst_compose.
eapply subst_ofc; eauto.
apply subof_dof_sof.
apply ctxsub_subof; auto.
}

(* tm *)
{
intros G2 i s' G3 t2 t3 _ IH Hindex Heqt23 H12.
simp_sub.
so (ctxsub_lookup_tm _#5 H12 Hindex) as (j & t1 & Heq & Hindex' & Heqt12).
rewrite -> Heq.
eapply ctxsub_tm; eauto.
eapply eqc_trans; eauto.
so (subst_eqc _#6 (subof_dof_sof _#3 (ctxsub_subof _#3 H12)) Heqt23) as H.
simp_sub in H.
exact H.
}

(* md *)
{
intros G2 i s' G3 T2 T3 c2 c3 _ IH Hindex HsubT23 Heqc23 H12.
simp_sub.
so (ctxsub_lookup_md _#6 H12 Hindex) as (j & T1 & c1 & Heq & Hindex' & HsubT12 & Heqc12).
rewrite -> Heq.
eapply ctxsub_md; eauto.
  {
  eapply subs_trans; eauto.
  so (subst_subs _#5 (subof_dof_sof _#3 (ctxsub_subof _#3 H12)) HsubT23) as H.
  simp_sub in H.
  exact H.
  }

  {
  etransitivity; eauto.
  rewrite <- compose_sh.
  rewrite -> !subst_compose.
  f_equal; auto.
  }
}
Qed.


(* If it's only constructor bindings, we don't need well-formedness. *)
Lemma ctxsub_append_simple :
  forall G1 s G2 G,
    ctxsub G1 s G2
    -> Forall (fun A => exists k, A = cl_cn k) G
    -> ctxsub (G1;; subst_ctx s G) (under (length G) s) (G2;; G).
Proof.
intros G1 s G2 G Hs HG.
induct HG.

(* nil *)
{
simpl.
auto.
}

(* cons *)
{
intros A G Heq _ IH.
destruct Heq as (k & ->).
simpl length.
rewrite -> under_succ.
simpl app.
simp_sub.
apply ctxsub_cn.
  {
  so (ctxsub_weaken _#3 (nil; cl_cn (subst (under (length G) s) k)) IH) as H.
  simpl app in H.
  simpl length in H.
  exact H.
  }

  {
  rewrite -> subst_compose.
  apply ofc_var.
  apply index_0.
  }
}
Qed.


Lemma ctxsub_append :
  forall G1 s G2 G,
    ctxsub G1 s G2
    -> wfh G2 G
    -> ctxsub (G1;; subst_ctx s G) (under (length G) s) (G2;; G).
Proof.
intros G1 s G2 G Hs HG.
revert G1 s Hs.
induct HG.

(* nil *)
{
intros G G1 s Hs.
simpl.
auto.
}

(* cn *)
{
intros G2 G k _ IH Hofk G1 s Hs.
simpl length.
rewrite -> under_succ.
simpl app.
simp_sub.
apply ctxsub_cn.
  {
  so (ctxsub_weaken _#3 (nil; cl_cn (subst (under (length G) s) k)) (IH _ _ Hs)) as H.
  simpl app in H.
  simpl length in H.
  exact H.
  }

  {
  rewrite -> subst_compose.
  apply ofc_var.
  apply index_0.
  }
}

(* tm *)
{
intros G2 G t Hwf IH Hoft G1 s Hs.
simpl length.
rewrite -> under_succ.
simpl app.
simp_sub.
eapply ctxsub_tm; eauto using index_0.
  {
  so (ctxsub_weaken _#3 (nil; cl_tm (subst (under (length G) s) t)) (IH _ _ Hs)) as H.
  simpl app in H.
  simpl length in H.
  exact H.
  }

  {
  simp_sub.
  apply eqc_refl.
  eapply subst_ofc_type; eauto.
  apply (subof_sof_trans _ (G1;; subst_ctx s G)).
    {
    apply subof_sh.
    auto using truncate_0, truncate_S.
    }

    {
    apply subof_under_sof.
      {
      eapply wfh_allcl; eauto.
      }

      {
      apply subof_dof_sof.
      apply ctxsub_subof; auto.
      }
    }
  }
}

(* md *)
{
intros G2 G k T Hwf IH HofT Hfst G1 s Hs.
simpl length.
rewrite <- !app_comm_cons.
rewrite -> !under_succ.
simp_sub.
rewrite -> !subst_ctx_cons.
rewrite <- !app_comm_cons.
simpl length.
simp_sub.
eapply ctxsub_md; eauto using index_0.
  {
  apply ctxsub_cn.
    {
    simpl length.
    simp_sub.
    so (ctxsub_weaken _#3 (nil; cl_cn (subst (under (length G) s) k); cl_md (subst (compose (under (length G) s) sh1) T) (var 0)) (IH _ _ Hs)) as H.
    simpl app in H.
    simpl length in H.
    exact H.
    }

    {
    simp_sub.
    eapply rel_quest.
      {
      eapply ofc_var; eauto using index_0, index_S.
      }
    simp_sub.
    reflexivity.
    }
  }

  {
  simp_sub.
  apply subs_refl.
  apply eqs_refl.
  eapply subst_ofs; eauto.
  apply (subof_sof_trans _ (G1;; subst_ctx s G)).
    {
    apply subof_sh.
    auto using truncate_0, truncate_S.
    }

    {
    apply subof_under_sof.
      {
      eapply wfh_allcl; eauto.
      }

      {
      apply subof_dof_sof.
      apply ctxsub_subof; auto.
      }
    }
  }
}
Qed.


Lemma ctxsub_refl :
  forall G,
    wf G
    -> ctxsub G (under (length G) id) G.
Proof.
intros G HG.
so (ctxsub_append _#4 (@ctxsub_nil nil) (wf_wfh _ HG)) as H.
rewrite -> !app_nil_r in H.
simpl length in H.
rewrite -> subst_ctx_id in H.
exact H.
Qed.


Lemma linv_allcn :
  forall P G T,
    linv P G T
    -> Forall (fun A => exists k, A = cl_cn k) G.
Proof.
intros P G T H.
induct H; eauto using Forall_nil, Forall_cons, Forall_app.
Qed.


Lemma linv_eqps :
  forall G P1 P2 G1 T1,
    wf G
    -> linv P1 G1 T1
    -> eqps G P2 P1
    -> exists G2 T2,
         linv P2 G2 T2
         /\ ctxsub (G;; G2) (under (length G + length G1) id) (G;; G1)
         /\ eqs (G;; G2) T2 T1
         /\ length G1 = length G2.
Proof.
intros G P1 P2 G1 T1 Hwf Hinv HeqP.
revert G1 T1 Hwf Hinv.
induct HeqP.

(* down *)
{
intros G T T' H G1 T1 Hwf Hinv.
invert Hinv; [].
intros <- <-.
exists nil, T.
do2 3 split; auto using linv_down.
simpl app.
simpl length.
replace (length G + 0) with (length G) by omega.
apply ctxsub_refl; auto.
}

(* ex *)
{
intros G k k' P P' Heqk HeqP IH GG T1 Hwf HinvGG.
invert HinvGG; [].
intros G1 Hinv <-.
so (eqk_reg _#3 Hwf Heqk) as (Hofk & _).
so (IH _ _ (wf_cn Hwf Hofk) Hinv) as (G2 & T2 & Hinv' & Hsub & HeqT & Hlen).
exists (nil; cl_cn k;; G2), T2.
do2 3 split.
  {
  apply linv_ex; auto.
  }

  {
  rewrite <- !app_assoc.
  simpl app.
  rewrite -> app_length.
  simpl length.
  simpl length in Hsub.
  replace (length G + (length G1 + 1)) with (1 + length G + length G1) by omega.
  replace (S (length G) + (length G1)) with (1 + length G + length G1) in Hsub by omega.
  replace (under (1 + length G + length G1) id) with (compose (under (1 + length G + length G1) id) (under (1 + length G + length G1) id)).
    {
    eapply ctxsub_trans; eauto.
    replace (1 + length G + length G1) with (length G1 + (1 + length G)) by omega.
    rewrite -> under_plus.
    replace (G; cl_cn k;; G1) with (G; cl_cn k;; subst_ctx (under (1 + length G) id) G1).
      {
      eapply ctxsub_append_simple; eauto using linv_allcn.
      rewrite -> under_succ.
      apply ctxsub_cn.
        {
        replace (G; cl_cn k) with (G;; (nil; cl_cn k)) by reflexivity.
        apply ctxsub_weaken.
        apply ctxsub_refl; auto.
        }

        {
        eapply ofc_subsume; eauto using ofc_var, index_0.
        apply subk_refl.
        simp_sub.
        rewrite -> subst_compose.
        simp_sub.
        eapply subst_eqk; eauto.
        apply weaken1.
        }
      }

      {
      f_equal.
      transitivity (subst_ctx id G1); auto using subst_ctx_id.
      apply subst_ctx_eqsub.
      apply eqsub_under_id.
      }
    }

    {
    rewrite -> dist_compose_under.
    reflexivity.
    }
  }

  {
  rewrite <- app_assoc.
  simpl app.
  assumption.
  }

  {
  rewrite -> !app_length.
  simpl length.
  omega.
  }
}
Qed.


Lemma linv_subps :
  forall G P1 P2 G1 T1,
    wf G
    -> linv P1 G1 T1
    -> subps G P2 P1
    -> exists s G2 T2,
         linv P2 G2 T2
         /\ ctxsub (G;; G2) s (G;; G1)
         /\ subs (G;; G2) T2 (subst s T1)
         /\ eqsub (compose (sh (length G1)) s) (sh (length G2)).
Proof.
intros G P1 P2 G1 T1 Hwf Hinv Hsubs.
revert G1 T1 Hwf Hinv.
induct Hsubs.

(* refl *)
{
intros G P P' HeqP G1 T1 Hwf Hinv.
so (linv_eqps _#5 Hwf Hinv HeqP) as (G2 & T2 & Hinv' & Hs & HeqT & Hlen).
exists (under (length G + length G1) id), G2, T2.
do2 3 split; auto.
  {
  rewrite -> subst_under_id.
  apply subs_refl; auto.
  }

  {
  rewrite -> compose_sh_under_leq; [| omega].
  replace (length G + length G1 - length G1) with (length G) by omega.
  rewrite <- Hlen.
  eapply (eqsub_trans _ (compose id (sh (length G1)))).
    {
    apply eqsub_compose; auto using eqsub_refl.
    apply eqsub_under_id.
    }

    {
    simp_sub.
    apply eqsub_refl.
    }
  }
}

(* trans *)
{
intros G P3 P2 P1 _ IH1 _ IH2 G1 T1 Hwf Hinv1.
so (IH2 _ _ Hwf Hinv1) as (s21 & G2 & T2 & Hinv2 & Hs21 & Hsubs21 & Heq21).
so (IH1 _ _ Hwf Hinv2) as (s32 & G3 & T3 & Hinv3 & Hs32 & Hsubs32 & Heq32).
exists (compose s21 s32), G3, T3.
do2 3 split; auto.
  {
  eapply ctxsub_trans; eauto.
  }

  {
  eapply subs_trans; eauto.
  rewrite -> subst_compose.
  eapply subst_subs; eauto.
  apply subof_dof_sof.
  apply ctxsub_subof; auto.
  }

  {
  rewrite <- compose_assoc.
  eapply eqsub_trans; eauto.
  apply eqsub_compose; auto using eqsub_refl.
  }
}

(* down *)
{
intros G T T' Hsub G1 T1 Hwf Hinv.
invert Hinv; [].
intros <- <-.
exists (under (length G) id), nil, T.
do2 3 split.
  {
  apply linv_down.
  }

  {
  simpl app.
  apply ctxsub_refl; auto.
  }

  {
  simpl app.
  simp_sub.
  auto.
  }

  {
  simpl length.
  simp_sub.
  apply eqsub_under_id.
  }
}

(* ex_left *)
{
intros G k P P' Hofk HsubP IH HofP' G1 T1 Hwf Hinv.
so (IH _ _ (wf_cn Hwf Hofk) (subst_linv _#4 Hinv)) as (s & G2 & T2 & Hinv' & Hs & HsubT & Heqs).
so (linv_reg_gen _#4 Hwf HofP' Hinv) as (Hwf1 & HofT1).
exists (compose (under (length G1) (compose (under (length G) id) sh1)) s), (nil; cl_cn k;; G2), T2.
assert (ctxsub (G; cl_cn k;; subst_ctx sh1 G1) (under (length G1) (compose (under (length G) id) sh1)) (G;; G1)) as Hctxsub.
  {
  replace (subst_ctx sh1 G1) with (subst_ctx (compose (under (length G) id) sh1) G1).
    {
    apply ctxsub_append; auto.
    so (ctxsub_weaken _#3 (nil; cl_cn k) (ctxsub_refl _ Hwf)) as H.
    simpl app in H.
    simpl length in H.
    exact H.
    }

    {
    apply subst_ctx_eqsub.
    apply (eqsub_trans _ (compose id sh1)).
      {
      apply eqsub_compose; auto using eqsub_refl.
      apply eqsub_under_id.
      }

      {
      simp_sub.
      apply eqsub_refl.
      }
    }
  }
do2 3 split.
  {
  apply linv_ex; auto.
  }

  {
  rewrite <- app_assoc.
  simpl app.
  eapply ctxsub_trans; eauto.
  }

  {
  rewrite <- app_assoc.
  simpl app.
  eapply subs_trans; eauto.
  simp_sub.
  eapply subs_refl.
  apply eqs_symm.
  eapply rel_quest.
    {
    eapply eqs_refl.
    eapply subst_ofs; eauto.
    eapply subof_sof_trans; eauto using ctxsub_subof, subof_dof_sof.
    }

    {
    apply subst_eqsub.
    apply eqsub_compose; auto using eqsub_refl.
    apply eqsub_under.
    apply (eqsub_trans _ (compose id sh1)).
      {
      apply eqsub_compose; auto using eqsub_refl.
      apply eqsub_under_id.
      }

      {
      simp_sub.
      apply eqsub_refl.
      }
    }
  }

  {
  rewrite -> app_length.
  simpl length.
  rewrite -> subst_ctx_length in Heqs.
  rewrite <- compose_assoc.
  rewrite -> compose_sh_under_leq; [| omega].
  replace (length G1 - length G1) with 0 by omega.
  rewrite -> under_zero.
  rewrite -> !compose_assoc.
  apply (eqsub_trans _ (compose id (sh (length G2 + 1)))).
    {
    apply eqsub_compose.
      {
      apply eqsub_under_id.
      }

      {
      replace (length G2 + 1) with (1 + length G2) by omega.
      rewrite <- compose_sh_sh.
      apply eqsub_compose; auto using eqsub_refl.
      }
    }

    {
    simp_sub.
    apply eqsub_refl.
    }
  }
}

(* ex_right *)
{
intros G P k P' c Hofc IsubP IH HofP' GG T1 Hwf HinvGG.
invert HinvGG; [].
intros G1 Hinv <-.
so (IH _ _ Hwf (subst_linv _#4 Hinv)) as (s & G2 & T2 & Hinv' & Hsub & HsubT & Heqs).
so (ofc_reg _#3 Hwf Hofc) as Hofk.
so (linv_reg_gen _#4 (wf_cn Hwf Hofk) HofP' Hinv) as (HwfG1 & HofT1).
exists (compose (under (length G1) (dot c (under (length G) id))) s), G2, T2.
do2 3 split; auto.
  {
  eapply ctxsub_trans; eauto.
  rewrite <- app_assoc.
  simpl app.
  replace (subst_ctx (dot c id) G1) with (subst_ctx (dot c (under (length G) id)) G1).
    {
    apply ctxsub_append; auto.
    apply ctxsub_cn.
      {
      apply ctxsub_refl; auto.
      }

      {
      simp_sub.
      assumption.
      }
    }
    
    {
    apply subst_ctx_eqsub.
    apply eqsub_dot.
    apply eqsub_under_id.
    }
  }

  {
  eapply subs_trans; eauto.
  simp_sub.
  apply subs_refl.
  eapply rel_quest.
    {
    eapply eqs_refl.
    eapply subst_ofs; eauto.
    eapply subof_sof_trans; eauto using ctxsub_subof, subof_dof_sof.
    apply subof_under_sof; eauto using wfh_allcl.
    apply subof_dot; auto using subof_id.
    simp_sub.
    apply sof_cn; auto.
    }

    {
    symmetry.
    apply subst_eqsub.
    apply eqsub_compose; auto using eqsub_refl.
    apply eqsub_under.
    apply eqsub_dot.
    apply eqsub_under_id.
    }
  }

  {
  rewrite -> app_length.
  simpl length.
  rewrite <- compose_assoc.
  rewrite -> compose_sh_under_geq; [| omega].
  replace (length G1 + 1 - length G1) with 1 by omega.
  simp_sub.
  rewrite -> subst_ctx_length in Heqs.
  apply (eqsub_trans _ (compose id (sh (length G2)))).
    {
    apply eqsub_compose; auto.
    apply eqsub_under_id.
    }

    {
    simp_sub.
    apply eqsub_refl.
    }
  }
}
Qed.


Lemma can_lclose :
  forall G P G' T P1,
    wf G
    -> ofps G P
    -> ofps (G;; G') P1
    -> linv P G' T
    -> exists P2,
         lclose P P1 P2
         /\ forall P3,
              subps (G;; G') P1 (subst (sh (length G')) P3)
              -> subps G P2 P3.
Proof.                                
intros G P G' T P1 Hwf HofP HofP1 Hinv.
revert G' T P1 Hwf HofP1 Hinv.
induct HofP.

(* down *)
{
intros G T HofT G' T' P1 _ HofP1 Hinv.
invert Hinv; [].
intros <- <-.
exists P1.
split; auto using lclose_down.
intros P3 Hsubs.
simp_sub in Hsubs.
exact Hsubs.
}

(* ex *)
{
intros G k P Hofk HofP IH GG T P1 Hwf HofP1 Hinv.
invert Hinv; [].
intros G' Hinv' <-.
rewrite <- app_assoc in HofP1.
simpl app in HofP1.
so (IH _#3 (wf_cn Hwf Hofk) HofP1 Hinv') as (P2 & Hclose & Hbest).
exists (ps_ex k P2).
split; auto using lclose_ex.
intros P3 Hsubs.
rewrite -> app_length in Hsubs.
simpl length in Hsubs.
rewrite <- app_assoc in Hsubs.
simpl app in Hsubs.
simp_sub in Hsubs.
apply subps_ex_left; auto.
  {
  apply Hbest; auto.
  simp_sub.
  exact Hsubs.
  }
so (linv_reg_gen _#4 (wf_cn Hwf Hofk) HofP Hinv') as (HwfG' & _).
so (subps_reg _#3 (wf_wfh_append _ _ (wf_cn Hwf Hofk) HwfG') Hsubs) as (_ & HofP3w).
so (wfh_inh _ _ HwfG') as (s & Hs & Hcancel).
replace P3 with (subst (compose s (dot (inh k) id)) (subst (sh (1 + length G')) P3)).
  {
  eapply subst_ofps; eauto.
  eapply subof_sof_trans; eauto.
  apply subof_dot; auto using subof_id.
  simp_sub.
  apply sof_cn.
  apply inh_ofc; auto.
  }

  {
  simp_sub.
  rewrite <- compose_sh_sh.
  rewrite -> compose_assoc.
  setoid_rewrite <- compose_assoc at 2.
  rewrite -> Hcancel.
  simp_sub.
  reflexivity.
  }
}
Qed.


Lemma prin_complete_gen :
  (forall G s G' e t',
     wf G
     -> ctxsub G s G'
     -> oft G' e t'
     -> exists t,
          print G (subst s e) t
          /\ eqc G t (subst s t') kd_type)
  /\
  (forall G s G' M T',
     wf G
     -> ctxsub G s G'
     -> ofm G' M T'
     -> exists T,
          prins G (subst s M) T
          /\ subs G T (subst s T'))
  /\
  (forall G s G' M T',
     wf G
     -> ctxsub G s G'
     -> oflm G' M T'
     -> exists T,
          prinps G (subst s M) T
          /\ subps G T (subst s T')).
Proof.
exploit
  (dynamic_ind
     (fun G' e t' =>
        forall s G,
          wf G
          -> ctxsub G s G'
          -> exists t,
               print G (subst s e) t
               /\ eqc G t (subst s t') kd_type)
     (fun G' M T' =>
        forall s G,
          wf G
          -> ctxsub G s G'
          -> exists T,
               prins G (subst s M) T
               /\ subs G T (subst s T'))
     (fun G' M T' =>
        forall s G,
          wf G
          -> ctxsub G s G'
          -> exists T,
               prinps G (subst s M) T
               /\ subps G T (subst s T'))) as Hind.

(* var *)
{
intros G' i t' Hindex s G _ Hs.
so (ctxsub_lookup_tm _#5 Hs Hindex) as (j & t & Heq & Hindex' & Heqt).
exists (subst (sh (S j)) t).
simp_sub.
rewrite -> Heq.
split.
  {
  apply print_var; auto.
  }
exact Heqt.
}

(* unit *)
{
intros G G' s Hwf Hs.
eauto using print with static; done.
}

(* lam *)
{
intros G t1 t2 e Hoft1 _ IH s G' Hwf Hs.
so (subst_ofc _#5 (subof_dof_sof _#3 (ctxsub_subof _ _ _ Hs)) Hoft1) as Hoft1'.
simp_sub in Hoft1'.
exploit (IH (under 1 s) (G'; cl_tm (subst s t1))) as (t2' & Hprint & Heqc).
  {
  apply wf_tm; auto.
  }

  {
  simp_sub.
  eapply ctxsub_tm; eauto with sequence.
    {
    so (ctxsub_weaken _#3 (nil; cl_tm (subst s t1)) Hs) as Hs'.
    exact Hs'.
    }

    {
    simp_sub.
    apply eqc_refl.
    rewrite -> subst_compose.
    apply weaken1_ofc_type; auto.
    }
  }
exists (cn_arrow (subst s t1) (subst1 nonsense t2')).
simp_sub.
simp_sub in Hprint.
split; eauto using print.
so (subst_eqc _#6 (unshift_tm _ _) Heqc) as Heqc'.
simp_sub in Heqc'.
eauto with static; done.
}

(* app *)
{
intros G t1 t2 e1 e2 _ IH1 _ IH2 s G' Hwf Hs.
so (IH1 _ _ Hwf Hs) as (t & Hprint & Heqc).
simp_sub in Heqc.
so (arrow_headnorm _#4 Hwf Heqc) as (t1' & t2' & Hnorm & Heqc1 & Heqc2).
so (IH2 _ _ Hwf Hs) as (u & Hprint2 & Hequ).
exists t2'.
simp_sub.
split; eauto using print with dynamic static.
}

(* pair *)
{
intros G t1 t2 e1 e2 _ IH1 _ IH2 s G' Hwf Hs.
so (IH1 _ _ Hwf Hs) as (t1' & Hprint1 & Heqc1).
so (IH2 _ _ Hwf Hs) as (t2' & Hprint2 & Heqc2).
exists (cn_prod t1' t2').
simp_sub.
eauto using print with static; done.
}

(* pi1 *)
{
intros G t1 t2 e _ IH s G' Hwf Hs.
so (IH _ _ Hwf Hs) as (t & Hprint & Heqc).
simp_sub in Heqc.
so (prod_headnorm _#4 Hwf Heqc) as (t1' & t2' & Hnorm & Heqc1 & Heqc2).
exists t1'.
simp_sub.
eauto using print with dynamic static; done.
}

(* pi2 *)
{
intros G t1 t2 e _ IH s G' Hwf Hs.
so (IH _ _ Hwf Hs) as (t & Hprint & Heqc).
simp_sub in Heqc.
so (prod_headnorm _#4 Hwf Heqc) as (t1' & t2' & Hnorm & Heqc1 & Heqc2).
exists t2'.
simp_sub.
eauto using print with dynamic static; done.
}

(* plam *)
{
intros G k t e Hofk _ IH Hval s G' Hwf Hs.
so (subst_ofk _#4 (subof_dof_sof _#3 (ctxsub_subof _ _ _ Hs)) Hofk) as Hofk'.
simp_sub in Hofk'.
exploit (IH (under 1 s) (G'; cl_cn (subst s k))) as (t' & Hprint & Heqc).
  {
  apply wf_cn; auto.
  }

  {
  simp_sub.
  apply ctxsub_cn.
    {
    so (ctxsub_weaken _#3 (nil; cl_cn (subst s k)) Hs) as Hs'.
    exact Hs'.
    }

    {
    rewrite -> subst_compose.
    apply ofc_var; eauto with sequence.
    }
  }
exists (cn_all (subst s k) t').
simp_sub.
simp_sub in Hprint.
split.
  {
  apply print_plam; auto.
  apply subst_svalue; auto.
  }

  {
  simp_sub in Heqc.
  apply eqc_all; auto using eqk_refl.
  }
}

(* papp *)
{
intros G k t e c _ IH Hofc s G' Hwf Hs.
so (IH _ _ Hwf Hs) as (u & Hprint & Heqc).
simp_sub in Heqc.
so (all_headnorm _#4 Hwf Heqc) as (k' & t' & Hnorm & Heqk & Heqc').
so (subst_ofc _#5 (subof_dof_sof _#3 (ctxsub_subof _ _ _ Hs)) Hofc) as Hofc'.
simp_sub in Hofc'.
exists (subst1 (subst s c) t').
simp_sub.
split.
  {
  eauto using print with static; done.
  }

  {
  replsub kd_type with (subst1 (subst s c) kd_type).
  rewrite -> split_dot.
  rewrite -> subst_compose.
  eapply subst1_eqc; eauto; done.
  }
}

(* pack *)
{
intros G k t c e Hofc _ IH Hoft s G' Hwf Hs.
exists (subst s (cn_exists k t)).
so (subst_ofc _#5 (subof_dof_sof _#3 (ctxsub_subof _ _ _ Hs)) Hofc) as Hofc'.
simp_sub in Hofc'.
so (IH _ _ Hwf Hs) as (u & Hprint & Heqc).
so (ofc_reg _#3 Hwf Hofc') as Hofk.
simp_sub.
split.
  {
  eapply print_pack; eauto.
    {
    simp_sub.
    simp_sub in Heqc.
    exact Heqc.
    }

    {
    eapply subst_ofc_type; eauto.
    apply subof_bind_sof_cn.
    apply subof_dof_sof.
    apply ctxsub_subof; auto.
    }
  }

  {
  apply eqc_refl.
  apply ofc_exists; auto.
  eapply subst_ofc_type; eauto.
  apply subof_bind_sof_cn.
  apply subof_dof_sof.
  apply ctxsub_subof; auto.
  }
}

(* unpack *)
{
intros G k t v e1 e2 _ IH1 _ IH2 s G' Hwf Hs.
so (IH1 _ _ Hwf Hs) as (u1 & Hprint1 & Heqc1).
simp_sub in Heqc1.
so (exists_headnorm _#4 Hwf Heqc1) as (k' & t' & Hnorm & Heqk & Heqt).
assert (ctxsub (G'; cl_cn k'; cl_tm t') (under 2 s) (G; cl_cn k; cl_tm t)) as Hs'.
  {
  rewrite -> !under_succ, -> under_zero.
  simp_sub.
  eapply ctxsub_tm; eauto using index_0.
    {
    apply ctxsub_cn; eauto using index_S, index_0.
      {
      replace (G'; cl_cn k'; cl_tm t') with (G';; (nil; cl_cn k'; cl_tm t')) by reflexivity.
      apply ctxsub_weaken; auto.
      }
    eapply ofc_subsume.
      {
      eapply ofc_var; eauto using index_0, index_S.
      }
    apply subk_refl.
    rewrite -> subst_compose.
    eapply subst_eqk; eauto using weaken2.
    }
  
    {
    eassert Prop as H; [refine (subst_eqc_type (G'; cl_cn k'; cl_tm t') _ (dot (var 1) (sh 2)) _ _ _ Heqt) |].
      {
      apply subof_dot; eauto using subof_sh, truncate_S, truncate_0.
      simp_sub.
      apply sof_cn.
      rewrite -> subst_compose.
      eapply ofc_subsume.
        {
        eapply ofc_var; eauto using index_S, index_0.
        }
      apply subk_refl.
      eapply subst_eqk; eauto using weaken2.
      }
    simp_sub in H.
    rewrite -> subst_dot1_sh in H.
    exact H.
    }
  }
so (eqk_reg _#3 Hwf Heqk) as (Hofk' & _).
assert (wf (G'; cl_cn k'; cl_tm t')) as Hwf'.
  {
  apply wf_tm; auto using wf.
  so (eqc_reg _#4 (wf_cn Hwf Hofk') (subst_eqc _#6 (subof_conv_0 _#3 Heqk) Heqt)) as (Hoft' & _).
  simp_sub in Hoft'.
  exact Hoft'.
  }
so (IH2 _ _ Hwf' Hs') as (u2 & Hprint2 & Heqc2).
exists (subst (dot nonsense (dot (inh k') id)) u2).
simp_sub.
so (subst_eqc _#6 (unshift_tm _ _) Heqc2) as Heqc2'.
simp_sub in Heqc2'.
so (subst1_eqc _#6 (inh_ofc _ _ Hofk') Heqc2') as Heqc2''.
simp_sub in Heqc2''.
split; auto; [].
simp_sub in Hprint2.
eapply print_unpack; eauto.
eapply eqc_trans; eauto; [].
apply eqc_symm; [].
so (weaken1_eqc _ (cl_cn k') _#3 Heqc2'') as Heqc2'''.
simp_sub in Heqc2'''.
assumption.
}

(* fix *)
{
intros G t e _ IH s G' Hwf Hs.
so (IH _ _ Hwf Hs) as (u & Hprint & Heqc).
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc anderl)) as (l & Hofcr).
invert Hofcr; [].
intros _ Hoft _.
simp_sub.
eauto using print with static; done.
}

(* lett *)
{
intros G t1 t2 e1 e2 _ IH1 _ IH2 s G' Hwf Hs.
so (IH1 _ _ Hwf Hs) as (t1' & Hprint1 & Heqc1).
so (eqc_reg _#4 Hwf Heqc1) as (Hoft1' & _).
exploit (IH2 (under 1 s) (G'; cl_tm t1')) as (t2' & Hprint2 & Heqc2).
  {
  apply wf_tm; auto.
  }

  {
  simp_sub.
  eapply ctxsub_tm; eauto with sequence.
    {
    so (ctxsub_weaken _#3 (nil; cl_tm t1') Hs) as Hs'.
    exact Hs'.
    }
  rewrite -> subst_compose.
  apply weaken1_eqc_type; auto.
  }
exists (subst1 nonsense t2').
simp_sub.
simp_sub in Hprint2.
split; eauto using print; [].
so (subst_eqc _#6 (unshift_tm _ _) Heqc2) as Heqc2'.
simp_sub in Heqc2'.
assumption.
}

(* equiv *)
{
intros G t t' e _ IH Heqc s G' Hwf Hs.
so (IH _ _ Hwf Hs) as (u & Hprint & Heqcu).
simp_sub in Heqc.
so (subst_eqc _#6 (subof_dof_sof _#3 (ctxsub_subof _ _ _ Hs)) Heqc) as Heqc'.
simp_sub in Heqc'.
eauto using print with static; done.
}

(* letm *)
{
intros G' P' Gi' T' k' t' M e _ IH1 Hinv Hfst' _ IH2 s G Hwf Hs.
so (IH1 _ _ Hwf Hs) as (P & Hprin1 & Hsub1).
so (subps_reg _#3 Hwf Hsub1) as (HofP & _).
so (subst_linv s _ _ _ Hinv) as Hinv'.
so (linv_subps _#5 Hwf Hinv' Hsub1) as (si & Gi & T & Hinv'' & Hsi & HsubsT & Heqsi).
rewrite -> subst_ctx_length in Heqsi.
so (linv_reg_gen _#4 Hwf HofP Hinv'') as (Hwfih & HofT).
so (wf_wfh_append _ _ Hwf Hwfih) as Hwfi.
so (can_fsts _ _ HofT) as (k & Hfst).
so (wf_md Hwfi HofT Hfst) as HwfiT.
so (ofs_fst _#3 HofT Hfst) as Hofk.
assert (ctxsub (G;; Gi; cl_cn k; cl_md (shift1 T) (var 0)) (under 2 (compose (under (length Gi') s) si)) (G';; Gi'; cl_cn k'; cl_md (shift1 T') (var 0))) as Hss.
  {
  simp_sub.
  simp_sub in HsubsT.
  eapply ctxsub_md; eauto using index_0.
    {
    apply ctxsub_cn.
      {
      rewrite <- compose_assoc.
      replace (G;; Gi; cl_cn k; cl_md (shift1 T) (var 0)) with ((G;; Gi);; (nil; cl_cn k; cl_md (shift1 T) (var 0))) by (autorewrite with canonlist; reflexivity).
      apply ctxsub_weaken.
      eapply ctxsub_trans; eauto.
      apply ctxsub_append_simple; auto.
      eapply linv_allcn; eauto.
      }
    eapply ofc_subsume.
      {
      eapply ofc_var; eauto using index_0, index_S.
      }
    rewrite <- compose_assoc.
    rewrite -> subst_compose.
    so (subs_fst _#5 Hwfi HsubsT Hfst (subst_fsts _#3 Hfst')) as Hsubk.
    exact (subst_subk _#5 (weaken2 _#4) Hsubk).
    }
  simp_sub.
  rewrite <- compose_assoc.
  rewrite -> subst_compose.
  exact (subst_subs _#5 (weaken2 _#4) HsubsT).
  }
so (IH2 _ _ HwfiT Hss) as (t & Hprin2 & Heq2).
simp_sub in Heq2.
replace (2 + length Gi') with (S (S (length Gi'))) in Heq2 by omega.
simp_sub in Heq2.
rewrite <- compose_assoc in Heq2.
rewrite -> compose_sh_under_leq in Heq2; [| omega].
replace (length Gi' - length Gi') with 0 in Heq2 by omega.
simp_sub in Heq2.
setoid_rewrite <- compose_assoc in Heq2 at 2.
rewrite -> (subst_eqsub _ (compose s (sh (2 + length Gi))) t') in Heq2.
  {
  assert (eqc G (subst (dot nonsense (dot (subst (cinh Gi) (inh k)) (cinh Gi))) t) (subst s t') kd_type) as Heq2'.
    {
    eapply rel_quest_1.
      {
      eapply subst_eqc_type; eauto.
      apply subof_dot.
        {
        apply subof_dot; eauto using cinh_subof.
        simp_sub.
        apply sof_cn.
        eapply subst_ofc; eauto using cinh_subof.
        apply inh_ofc; auto.
        }
      simp_sub.
      apply sof_md.
      }
    simp_sub.
    replace (2 + length Gi) with (S (S (length Gi))) by omega.
    simp_sub.
    rewrite -> cinh_trunc.
    simp_sub.
    reflexivity.
    }
  exists (subst (dot nonsense (dot (subst (cinh Gi) (inh k)) (cinh Gi))) t).
  split; auto.
  simp_sub.  
  apply (@print_letm _ P Gi T _ t); auto.
    {
    simp_sub in Hprin2.
    replace (1 + length Gi') with (S (length Gi')) in Hprin2 by omega.
    simp_sub in Hprin2.
    rewrite <- compose_assoc in Hprin2.
    rewrite -> compose_sh_under_leq in Hprin2; [| omega].
    replace (length Gi' - length Gi') with 0 in Hprin2 by omega.
    simp_sub in Hprin2.
    setoid_rewrite <- compose_assoc in Hprin2 at 2.
    simp_sub.
    replace (2 + length Gi) with (length Gi + 2) by omega.
    rewrite <- compose_sh_sh.
    force_exact Hprin2.
    f_equal; [].
    apply subst_eqsub.
    apply eqsub_dot.
    apply eqsub_dot.
    apply eqsub_compose; auto using eqsub_refl.
    apply eqsub_compose; auto using eqsub_refl.
    }

    {
    apply (@eqc_trans _ _ (subst (compose s (sh (S (length Gi)))) t')).
      {
      eapply rel_quest_1.
        {
        eapply subst_eqc_type; eauto.
        apply subof_dot; auto using subof_id.
        simp_sub.
        apply sof_md.
        }
      simp_sub.
      replace (2 + length Gi) with (S (1 + length Gi)) by omega.
      simp_sub.
      reflexivity.
      }

      {
      apply eqc_symm.
      setoid_rewrite -> subst_compose at 2.
      eapply rel_quest_2.
        {
        eapply subst_eqc_type; eauto.
        rewrite -> app_comm_cons.
        replace (S (length Gi)) with (length (Gi; cl_cn k)) by reflexivity.
        apply subof_sh; eauto.
        apply app_truncate.
        }
      simp_sub.
      reflexivity.
      }
    }
  }

  {
  replace (2 + length Gi) with (length Gi + 2) by omega.
  rewrite <- compose_sh_sh.
  apply eqsub_compose; auto using eqsub_refl.
  apply eqsub_compose; auto using eqsub_refl.
  }
}

(* ext *)
{
intros G t M _ IH s G' Hwf Hs.
so (IH _ _ Hwf Hs) as (T & Hprins & Hsubs).
simp_sub in Hsubs.
so (subs_dyn_invert' _#3 Hsubs) as (t' & -> & Heqc).
simp_sub.
eauto using print with static; done.
}

(* var *)
{
intros G' i T' c' Hindex' s G HwfG Hs.
so (ctxsub_lookup_md _#6 Hs Hindex') as (j & T & c & Heq & Hindex & HsubT & Heqc).
exists (singles (subst (sh (S j)) c) (subst (sh (S j)) T)).
simp_sub.
split.
  {
  rewrite -> Heq.
  apply prins_var; auto.
  }
apply (@subs_trans _ _ (subst (sh (S j)) T)); auto.
so (wf_index_md _#4 HwfG Hindex) as (HofT & Heqcvar).
subst c.
so (can_fsts _ _ HofT) as (k & Hfst).
eapply singles_subs; eauto.
simp_sub.
so (ofm_fst _#5 HwfG (ofm_var Hindex) (fstm_var Hindex) Hfst) as H.
exact H.
}

(* star *)
{
intros G G' s _ Hs.
exists sg_one.
split; eauto using prins with static; done.
}

(* stat *)
{
intros G c k Hof s G' Hwf Hs.
so (subst_ofc _#5 (subof_dof_sof _#3 (ctxsub_subof _ _ _ Hs)) Hof) as Hof'.
simp_sub in Hof'.
so (prink_complete _#3 Hwf Hof') as (k' & Hprink & Hsubk).
exists (sg_stat k').
simp_sub.
split; eauto using prins with static.
}

(* dyn *)
{
intros G e t _ IH s G' Hwf Hs.
so (IH _ _ Hwf Hs) as (t' & Hprint & Heqc).
exists (sg_dyn t').
simp_sub.
split; eauto using prins with static; done.
}

(* lam *)
{
intros G T1 T2 k1 M HofT1 Hfst1 _ IH s G' Hwf Hs.
so (subst_ofs _#4 (subof_dof_sof _#3 (ctxsub_subof _ _ _ Hs)) HofT1) as HofT1'.
simp_sub in HofT1'.
assert (ctxsub (G'; cl_cn (subst s k1); cl_md (shift1 (subst s T1)) (var 0)) (under 2 s) (G; cl_cn k1; cl_md (shift1 T1) (var 0))) as Hs'.
  {
  rewrite -> !under_succ, -> under_zero; simp_sub.
  eapply ctxsub_md; eauto with sequence.
    {
    eapply ctxsub_cn.
      {
      so (ctxsub_weaken _#3 (nil; cl_cn (subst s k1); cl_md (shift1 (subst s T1)) (var 0)) Hs) as H.
      simp_sub in H.
      simpl app in H.
      simpl length in H.
      exact H.
      }

      {
      setoid_rewrite -> subst_compose at 2.
      apply ofc_var; eauto with sequence.
      }
    }

    {
    simp_sub.
    apply subs_refl.
    apply eqs_refl.
    setoid_rewrite -> subst_compose at 2 3.
    eapply subst_ofs; eauto using weaken2.
    }
  }
so (wf_md Hwf (subst_ofs _#4 (subof_dof_sof _#3 (ctxsub_subof _#3 Hs)) HofT1) (subst_fsts _#3 Hfst1)) as Hwf'.
so (IH _ _ Hwf' Hs') as (T2' & Hprins & Hsubs).
exists (sg_pi (subst s T1) (subst1 nonsense T2')).
simp_sub.
split.
  {
  eapply prins_lam; eauto using subst_fsts.
  simp_sub.
  so (subord_md_sg _#4 (subs_reg _#3 Hwf' Hsubs andel)) as Heq.
  rewrite <- Heq; [].
  simp_sub in Hprins.
  exact Hprins.
  }

  {
  eapply subs_pi; eauto using subst_fsts with static.
    {
    so (subst_subs _#5 (unshift_md _#3) Hsubs) as Hsubs'.
    simp_sub in Hsubs'.
    assumption.
    }

    {
    so (subst_ofs _#4 (unshift_md _#3) (subs_reg _#3 Hwf' Hsubs andel)).
    assumption.
    }
  }
}

(* appap *)
{
intros G T1 T2 M1 M2 c2 _ IH1 _ IH2 HfstM2 s G' Hwf Hs.
so (IH1 _ _ Hwf Hs) as (T & Hprin1 & Hsubspi).
simp_sub in Hsubspi.
so (subs_pi_invert' _#4 Hwf Hsubspi) as (T1' & T2' & k1 & -> & Hfst1 & Hsubs1 & Hsubs2).
so (IH2 _ _ Hwf Hs) as (U & Hprins2 & HsubsU).
so (subst_fstm _#5 (subof_dof_tof _#3 (ctxsub_subof _ _ _ Hs)) HfstM2) as HfstM2'.
simp_sub in HfstM2'.
exists (subst1 (subst s c2) T2').
simp_sub.
split; eauto using prins with static.
rewrite -> split_dot.
rewrite -> subst_compose.
eapply subst1_subs; eauto; [].
so (ofm_fst_o _#4 Hwf (prins_sound _#3 Hwf Hprins2) HfstM2') as (k & HfstU & Hofc2).
so (subs_fst _#5 Hwf HsubsU HfstU Hfst1) as Hsubk.
eapply ofc_subsume; eauto; done.
}

(* pair *)
{
intros G T1 T2 M1 M2 _ IH1 _ IH2 s G' Hwf Hs.
so (IH1 _ _ Hwf Hs) as (T1' & Hprin1 & Hsubs1).
so (IH2 _ _ Hwf Hs) as (T2' & Hprin2 & Hsubs2).
simp_sub.
exists (sg_sigma T1' (shift1 T2')).
split; eauto using prins.
so (subs_reg _#3 Hwf Hsubs1) as (HofT1' & HofT1).
so (can_fsts _ _ HofT1) as (k1 & Hfst1).
so (can_fsts _ _ HofT1') as (k1' & Hfst1').
eapply subs_sigma; eauto.
  {
  rewrite -> subst_compose.
  eapply subst_subs; eauto using weaken1; done.
  }

  {
  so (subs_reg _#3 Hwf Hsubs2 ander) as Hof2.
  rewrite -> subst_compose.
  eapply subst_ofs; eauto using weaken1; done.
  }
}

(* pi1 *)
{
intros G T1 T2 M _ IH s G' Hwf Hs.
so (IH _ _ Hwf Hs) as (T & Hprin & Hsubs).
simp_sub in Hsubs.
so (subs_sigma_invert' _#4 Hwf Hsubs) as (T1' & T2' & k1' & -> & HfstT1' & Hsubs1 & Hsubs2).
exists T1'.
simp_sub.
split; eauto using prins with static; done.
}

(* pi2 *)
{
intros G T1 T2 M c _ IH HfstM s G' Hwf Hs.
so (IH _ _ Hwf Hs) as (T & Hprin & Hsubs).
simp_sub in Hsubs.
so (subs_sigma_invert' _#4 Hwf Hsubs) as (T1' & T2' & k1' & -> & HfstT1' & Hsubs1 & Hsubs2).
so (subst_fstm _#5 (subof_dof_tof _#3 (ctxsub_subof _ _ _ Hs)) HfstM) as HfstM'.
simp_sub in HfstM'.
exists (subst1 (cn_pi1 (subst s c)) T2').
simp_sub.
split; eauto using prins; [].
rewrite -> split_dot.
rewrite -> subst_compose.
eapply subst1_subs; eauto; [].
so (prins_sound _#3 Hwf Hprin) as Hof'.
exact (ofm_fst _#5 Hwf (ofm_pi1 Hof') (fstm_pi1 HfstM') HfstT1').
}

(* lett *)
{
intros G t T e M _ IH1 _ IH2 s G' Hwf Hs.
so (IH1 _ _ Hwf Hs) as (t' & Hprint & Heqc).
so (eqc_reg _#4 Hwf Heqc) as (Hoft' & _).
assert (ctxsub (G'; cl_tm t') (under 1 s) (G; cl_tm t)) as Hs'.
  {
  simp_sub.
  eapply ctxsub_tm; eauto with sequence.
    {
    so (ctxsub_weaken _#3 (nil; cl_tm t') Hs) as Hs'.
    exact Hs'.
    }

    {
    rewrite -> subst_compose.
    apply weaken1_eqc_type; auto.
    }
  }
so (wf_tm Hwf Hoft') as Hwf'.
so (IH2 _ _ Hwf' Hs') as (T' & Hprins & Hsubs).
simp_sub in Hprins.
exists (subst1 nonsense T').
simp_sub.
split; eauto using prins with static; [].
so (subst_subs _#5 (unshift_tm _ _) Hsubs) as Hsubs'.
simp_sub in Hsubs'.
exact Hsubs'.
}

(* cir *)
{
intros G' M P' _ IH s G Hwf Hs.
so (IH _ _ Hwf Hs) as (P & Hprins & Hsub).
exists (sg_circ P).
simp_sub.
split; eauto using prins with static.
}

(* extstat *)
{
intros G' k k' M c _ IH HfstM Hofc s G Hwf Hs.
so (IH _ _ Hwf Hs) as (U & Hprins & Hsubs).
simp_sub in Hsubs.
so (subs_stat_invert' _#3 Hsubs) as (l & -> & _).
exists (sg_stat l).
split; auto; [].
simp_sub.
apply subs_stat; [].
so (subst_ofc _#5 (subof_dof_sof _#3 (ctxsub_subof _#3 Hs)) Hofc) as Hofc'.
so (subst_fstm _#5 (subof_dof_tof _#3 (ctxsub_subof _#3 Hs)) HfstM) as HfstM'.
so (prink_complete _#3 Hwf Hofc') as (l' & Hprink & Hsubk).
so (prins_prink_fst _#5 Hwf HfstM' Hprins Hprink) as Hfsts.
invert Hfsts; [].
intros <-.
exact Hsubk.
}

(* extpi *)
{
intros G T1 T2 T3 k1 M _ IH1 HfstT1 _ IH2 s G' Hwf Hs.
so (IH1 _ _ Hwf Hs) as (T & Hprins & Hsubs).
exists T.
split; auto; [].
simp_sub.
simp_sub in Hsubs.
so (subs_pi_invert' _#4 Hwf Hsubs) as (T1' & T2' & _ & -> & _ & Hsubs1 & _).
so (subs_reg _#3 Hwf Hsubs) as (Hofapp & _).
invert Hofapp; [].
intros k1' HofT1' HfstT1' HofT2'.
eapply subs_pi; eauto using subst_fsts.
so (subs_reg _#3 Hwf Hsubs1) as (HofT1 & _).
so (wf_md Hwf HofT1 (subst_fsts _#3 HfstT1)) as Hwf'.
exploit (IH2 (under 2 s) (G'; cl_cn (subst s k1); cl_md (shift1 (subst s T1)) (var 0))) as (U & Hprins' & Hsubs2); auto.
  {
  rewrite -> !under_succ, -> under_zero; simp_sub.
  eapply ctxsub_md; auto using index_0.
    {
    apply ctxsub_cn.
      {
      exact (ctxsub_weaken _#3 (nil; cl_cn (subst s k1); cl_md (subst (compose s sh1) T1) (var 0)) Hs).
      }
    setoid_rewrite -> subst_compose at 2.
    apply ofc_var; eauto with sequence.
    }
  
    {
    simp_sub.
    apply subs_refl.
    apply eqs_refl.
    setoid_rewrite -> subst_compose at 2.
    exact (subst_ofs _#4 (weaken2 _#4) HofT1).
    }
  }
cut (U = shift1 T2').
  {
  intros ->.
  so (subst_subs _#5 (unshift_md _#3) Hsubs2) as Hsubs2'.
  simp_sub in Hsubs2'.
  assumption.
  }
eapply (prins_fun _ _); eauto; [].
exploit (@prins_app (G'; cl_cn (subst s k1); cl_md (shift1 (subst s T1)) (var 0)) (subst (sh 2) T1') (subst (dot (var 0) (sh 3)) T2') (singles (shift1 (var 0)) (shift1 (shift1 (subst s T1)))) (subst (compose s (sh 2)) M) (var 0) (var 1)) as H.
  {
  so (weaken1_prins _#3 (cl_md (shift1 (subst s T1)) (var 0)) (weaken1_prins _#3 (cl_cn (subst s k1)) Hprins)) as Hprinsx.
  simp_sub in Hprinsx.
  simp_sub.
  exact Hprinsx.
  }

  {
  apply prins_var; eauto using index_0.
  }

  {
  simp_sub.
  eapply subs_trans; eauto using subst_subs, weaken2; [].
  simp_sub.
  simp_sub in Hwf'.
  eapply singles_subs; eauto using subst_fsts, subst_ofs, weaken2 with static sequence.
    {
    setoid_rewrite -> subst_compose at 2.
    eapply subst_ofs; eauto using weaken2.
    }

    {
    setoid_rewrite -> subst_compose at 2.
    eapply ofc_var; eauto with sequence.
    }
  }

  {
  replsub (var 1) with (shift1 (var 0)).
  eapply fstm_var; eauto with sequence; done.
  }
simp_sub in H.
rewrite -> subst_dot1_sh in H; [].
simp_sub.
exact H.
}

(* extsigma *)
{
intros G T1 T2 M _ IH1 _ IH2 s G' Hwf Hs.
so (IH1 _ _ Hwf Hs) as (U1 & Hprins1 & Hsubs1).
so (IH2 _ _ Hwf Hs) as (U2x & Hprins2 & Hsubs2).
invert Hprins1; [].
intros U2 Hprins.
invert Hprins2; [].
intros U1' U2' c Hprins' HfstM <-.
so (prins_fun _#4 Hprins Hprins') as Heq.
injection Heq.
intros <- <-.
so (prins_ndss _#3 Hprins) as Hndss.
invert Hndss; [].
intros U2' _ _ <-.
simp_sub in Hsubs2.
exists (sg_sigma U1 (shift1 U2')).
split; auto; [].
so (subs_reg _#3 Hwf Hsubs1) as (HofU1 & HofT1).
so (can_fsts _ _ HofU1) as (l & HfstU1).
so (can_fsts _ _ HofT1) as (k & HfstT1).
so (subs_reg _#3 Hwf Hsubs2) as (_ & HofU2').
simp_sub.
#2 eapply subs_sigma; eauto.
  {
  rewrite -> subst_compose.
  eapply subst_subs; eauto using weaken1; done.
  }

  {
  rewrite -> subst_compose.
  eapply subst_ofs; eauto using weaken1; done.
  }
}

(* subsume *)
{
intros G' T T' M _ IH HsubT s G Hwf Hs.
so (IH _ _ Hwf Hs) as (U & Hprins & HsubU).
exists U.
split; auto.
eapply subs_trans; eauto.
eapply subst_subs; eauto.
apply subof_dof_sof.
apply ctxsub_subof; auto.
}

(* ret *)
{
intros G' M T' _ IH s G Hwf Hs.
so (IH _ _ Hwf Hs) as (T & Hprins & Hsub).
exists (ps_down T).
simp_sub.
split; eauto using prinps with static.
}

(* seal *)
{
intros G' M T _ IH s G Hwf Hs.
so (IH _ _ Hwf Hs) as (U & Hprins & Hsub).
exists (ps_down (subst s T)).
split.
  {
  simp_sub.
  eapply prinps_seal; eauto.
  }

  {
  simp_sub.
  apply subps_refl.
  apply eqps_refl.
  apply ofps_down.
  exact (subs_reg _#3 Hwf Hsub ander).
  }
}

(* bind *)
{
intros G' M1 M2 P1' Gi' T' k' P2' HofM1 IH1 Hinv Hfst' HofM2 IH2 s G Hwf Hs.
so (IH1 _ _ Hwf Hs) as (T & Hprin1 & Hsubcirc).
simp_sub in Hsubcirc.
so (subs_circ_invert' _#3 Hsubcirc) as (P1 & -> & Hsubs1).
so (subps_reg _#3 Hwf Hsubs1) as (HofP1 & _).
so (subst_linv s _ _ _ Hinv) as Hinv'.
so (linv_subps _#5 Hwf Hinv' Hsubs1) as (si & Gi & T & Hinv'' & Hsi & HsubsT & Heqsi).
rewrite -> subst_ctx_length in Heqsi.
so (linv_reg_gen _#4 Hwf HofP1 Hinv'') as (Hwfih & HofT).
so (wf_wfh_append _ _ Hwf Hwfih) as Hwfi.
so (can_fsts _ _ HofT) as (k & Hfst).
so (wf_md Hwfi HofT Hfst) as HwfiT.
so (ofs_fst _#3 HofT Hfst) as Hofk.
assert (ctxsub (G;; Gi; cl_cn k; cl_md (shift1 T) (var 0)) (under 2 (compose (under (length Gi') s) si)) (G';; Gi'; cl_cn k'; cl_md (shift1 T') (var 0))) as Hss.
  {
  simp_sub.
  simp_sub in HsubsT.
  eapply ctxsub_md; eauto using index_0.
    {
    apply ctxsub_cn.
      {
      rewrite <- compose_assoc.
      replace (G;; Gi; cl_cn k; cl_md (shift1 T) (var 0)) with ((G;; Gi);; (nil; cl_cn k; cl_md (shift1 T) (var 0))) by (autorewrite with canonlist; reflexivity).
      apply ctxsub_weaken.
      eapply ctxsub_trans; eauto.
      apply ctxsub_append_simple; auto.
      eapply linv_allcn; eauto.
      }
    eapply ofc_subsume.
      {
      eapply ofc_var; eauto using index_0, index_S.
      }
    rewrite <- compose_assoc.
    rewrite -> subst_compose.
    so (subs_fst _#5 Hwfi HsubsT Hfst (subst_fsts _#3 Hfst')) as Hsubk.
    exact (subst_subk _#5 (weaken2 _#4) Hsubk).
    }
  simp_sub.
  rewrite <- compose_assoc.
  rewrite -> subst_compose.
  exact (subst_subs _#5 (weaken2 _#4) HsubsT).
  }
so (IH2 _ _ HwfiT Hss) as (P2 & Hprin2 & Hsub2).
so (subps_reg _#3 HwfiT Hsub2) as (HofP2 & HofP2').
assert (ofps (G;; Gi) (ps_ex k (subst1 nonsense P2))) as Hofex.
  {
  apply ofps_ex; auto.
  unfold subst1.
  eapply subst_ofps; eauto.
  apply subof_dot; auto using subof_id.
  simp_sub.
  apply sof_md.
  }
so (can_lclose _#5 Hwf HofP1 Hofex Hinv'') as (P3 & Hclose & Hbest).
exists P3.
split.
  {
  simp_sub.
  eapply prinps_bind; eauto.
  simp_sub.
  simp_sub in Hprin2.
  replace (1 + length Gi') with (S (length Gi')) in Hprin2 by omega.
  rewrite -> compose_sh_succ_dot in Hprin2.
  rewrite <- compose_assoc in Hprin2.
  rewrite -> compose_sh_under_leq in Hprin2; [| omega].
  replace (length Gi' - length Gi') with 0 in Hprin2 by omega.
  rewrite -> under_zero in Hprin2.
  rewrite -> compose_assoc in Hprin2.
  setoid_rewrite <- compose_assoc in Hprin2 at 2.
  rewrite -> (subst_eqsub _#3 (eqsub_dot _#3 (eqsub_dot _#3 (eqsub_compose _#4 (eqsub_refl _) (eqsub_compose _#4 Heqsi (eqsub_refl _)))))) in Hprin2.
  rewrite -> compose_sh_sh in Hprin2.
  replace (2 + length Gi) with (length Gi + 2) by omega.
  exact Hprin2.
  }
cut (subps (G;; Gi) (subst (sh (length Gi)) P3) (subst (compose s (sh (length Gi))) P2')).
  {
  intro Hsubs3.
  so (wfh_inh _ _ Hwfih) as (sc & Hsc & Hcancel).
  so (subst_subps _#5 Hsc Hsubs3) as H.
  rewrite <- !subst_compose in H.
  rewrite -> compose_assoc in H.
  rewrite -> !Hcancel in H.
  simp_sub in H.
  exact H.
  }
exploit (Hbest (subst s P2')) as H.
  {
  apply subps_ex_left; auto.
    {
    simp_sub.
    eassert Prop as H; [refine (subst_subps (G;; Gi; cl_cn k) _ (dot nonsense id) _ _ _ Hsub2) |].
      {
      apply subof_dot; auto using subof_id.
      simp_sub.
      apply sof_md.
      }
    replace (2 + length Gi') with (S (S (length Gi'))) in H by omega.
    simp_sub in H.
    rewrite <- compose_assoc in H.
    rewrite -> compose_sh_under_leq in H; [| omega].
    replace (length Gi' - length Gi') with 0 in H by omega.
    rewrite -> under_zero in H.
    rewrite -> compose_assoc in H.
    setoid_rewrite <- compose_assoc in H at 2.
    rewrite -> (subst_eqsub _#3 (eqsub_compose _#4 (eqsub_refl _) (eqsub_compose _#4 Heqsi (eqsub_refl _)))) in H.
    simp_sub in H.
    exact H.
    }
  
    {
    simp_sub in HofP2'.
    replace (2 + length Gi') with (S (S (length Gi'))) in HofP2' by omega.
    rewrite -> !compose_sh_succ_dot in HofP2'.
    rewrite <- compose_assoc in HofP2'.
    rewrite -> compose_sh_under_leq in HofP2'; [| omega].
    replace (length Gi' - length Gi') with 0 in HofP2' by omega.
    rewrite -> under_zero in HofP2'.
    rewrite -> !compose_assoc in HofP2'.
    setoid_rewrite <- compose_assoc in HofP2' at 2.
    rewrite -> (subst_eqsub _#3 (eqsub_compose _#4 (eqsub_refl _) (eqsub_compose _#4 Heqsi (eqsub_refl _)))) in HofP2'.
    eassert Prop as H; [refine (subst_ofps (G;; Gi) _ (dot nonsense (dot (inh k) id)) _ _ HofP2') |].
      {
      apply subof_dot.
        {
        apply subof_dot; auto using subof_id.
        simp_sub.
        apply sof_cn.
        apply inh_ofc; auto.
        }
      simp_sub.
      apply sof_md.
      }
    simp_sub in H.
    simp_sub.
    exact H.
    }
  }
rewrite -> subst_compose.
eapply subst_subps; eauto.
apply subof_sh.
apply app_truncate.
}

(* unpack *)
{
intros G' k' t' P e M _ IH1 _ IH2 s G Hwf Hs.
so (IH1 _ _ Hwf Hs) as (u & Hprint & Hequ).
simp_sub in Hequ.
so (exists_headnorm _#4 Hwf Hequ) as (k & t & Hnorm & Heqk & Heqt).
so (eqk_reg _#3 Hwf Heqk) as (Hofk & Hofk').
so (eqc_reg _#4 (wf_cn Hwf Hofk') Heqt) as (Hoftalt & _).
so (subst_ofc_type _#4 (subof_conv_0 _#3 Heqk) Hoftalt) as Hoft.
clear Hoftalt.
rewrite -> subst_under_id in Hoft.
assert (ctxsub (G; cl_cn k; cl_tm t) (under 2 s) (G'; cl_cn k'; cl_tm t')) as Hs'.
  {
  rewrite -> !under_succ, -> under_zero.
  simp_sub.
  eapply ctxsub_tm; eauto using index_0.
    {
    apply ctxsub_cn; eauto using index_S, index_0.
      {
      replace (G; cl_cn k; cl_tm t) with (G;; (nil; cl_cn k; cl_tm t)) by reflexivity.
      apply ctxsub_weaken; auto.
      }
    eapply ofc_subsume.
      {
      eapply ofc_var; eauto using index_0, index_S.
      }
    apply subk_refl.
    rewrite -> subst_compose.
    eapply subst_eqk; eauto using weaken2.
    }
  
    {
    eassert Prop as H; [refine (subst_eqc_type (G; cl_cn k; cl_tm t) _ (dot (var 1) (sh 2)) _ _ _ Heqt) |].
      {
      apply subof_dot; eauto using subof_sh, truncate_S, truncate_0.
      simp_sub.
      apply sof_cn.
      rewrite -> subst_compose.
      eapply ofc_subsume.
        {
        eapply ofc_var; eauto using index_S, index_0.
        }
      apply subk_refl.
      eapply subst_eqk; eauto using weaken2.
      }
    simp_sub in H.
    rewrite -> subst_dot1_sh in H.
    exact H.
    }
  }
so (wf_tm (wf_cn Hwf Hofk) Hoft) as Hwf'.
so (IH2 _ _ Hwf' Hs') as (T & Hprins & Hsubs).
exists (ps_ex k (subst1 nonsense T)).
simp_sub.
split.
  {
  eapply prinps_unpack; eauto.
  rewrite -> !under_succ, -> under_zero in Hprins.
  simp_sub in Hprins.
  exact Hprins.
  }

  {
  apply subps_ex_left; auto.
    {
    eassert Prop as H; [refine (subst_subps (G; cl_cn k) _ (dot nonsense id) _ _ _ Hsubs) |].
      {
      apply subof_dot; auto using subof_id.
      simp_sub.
      apply sof_tm.
      }
    simp_sub in H.
    simp_sub.
    exact H.
    }

    {
    simp_sub in Hsubs.
    so (subps_reg _#3 Hwf' Hsubs) as (_ & HofPw).
    replace (subst s P) with (subst (dot nonsense (dot (inh k) id)) (subst (compose s (sh 2)) P)) by (simp_sub; reflexivity).
    eapply subst_ofps; eauto.
    apply subof_dot.
      {
      apply subof_dot; auto using subof_id.
      simp_sub.
      apply sof_cn.
      apply inh_ofc; auto.
      }
    simp_sub.
    apply sof_tm.
    }
  }
}

(* subsume *)
{
intros G' P P' M _ IH HsubP s G Hwf Hs.
so (IH _ _ Hwf Hs) as (T & Hprins & HsubT).
exists T.
split; auto.
eapply subps_trans; eauto.
eapply subst_subps; eauto.
apply subof_dof_sof.
apply ctxsub_subof; auto.
}

(* epilogue *)
{
destruct_all Hind.
split; intros; eauto.
}
Qed.


Lemma print_complete :
  forall G e t,
    wf G
    -> oft G e t
    -> exists t',
         print G e t'
         /\ eqc G t' t kd_type.
Proof.
intros G e t Hwf Hof.
so (prin_complete_gen andel _#5 Hwf (ctxsub_refl _ Hwf) Hof) as H.
rewrite -> !subst_under_id in H.
exact H.
Qed.


Lemma prins_complete :
  forall G M T,
    wf G
    -> ofm G M T
    -> exists T',
         prins G M T'
         /\ subs G T' T.
Proof.
intros G M T Hwf Hof.
so (prin_complete_gen anderl _#5 Hwf (ctxsub_refl _ Hwf) Hof) as H.
rewrite -> !subst_under_id in H.
exact H.
Qed.


Lemma prinps_complete :
  forall G M P,
    wf G
    -> oflm G M P
    -> exists P',
         prinps G M P'
         /\ subps G P' P.
Proof.
intros G M T Hwf Hof.
so (prin_complete_gen anderr _#5 Hwf (ctxsub_refl _ Hwf) Hof) as H.
rewrite -> !subst_under_id in H.
exact H.
Qed.


Lemma oft_oftp :
  forall G e t,
    wf G
    -> oft G e t
    -> oftp G e t.
Proof.
intros G e t Hwf Hof.
so (print_complete _#3 Hwf Hof) as (t' & Hprin & Heqc).
so (print_oftp _#3 Hwf Hprin) as Hof'.
eapply oftp_equiv; eauto; done.
Qed.


Lemma ofm_ofmp :
  forall G M T,
    wf G
    -> ofm G M T
    -> ofmp G M T.
Proof.
intros G M T Hwf Hof.
so (prins_complete _#3 Hwf Hof) as (T' & Hprin & Hsubs).
so (prins_ofmp _#3 Hwf Hprin) as Hof'.
so (ofmp_subsume Hof' Hsubs) as Hof''.
exact Hof''.
Qed.


Lemma oflm_oflmp :
  forall G M P,
    wf G
    -> oflm G M P
    -> oflmp G M P.
Proof.
intros G M T Hwf Hof.
so (prinps_complete _#3 Hwf Hof) as (T' & Hprin & Hsubs).
so (prinps_oflmp _#3 Hwf Hprin) as Hof'.
so (oflmp_subsume Hof' Hsubs) as Hof''.
exact Hof''.
Qed.


Lemma oft_unique :
  forall G e t t',
    wf G
    -> oft G e t
    -> oft G e t'
    -> eqc G t t' kd_type.
Proof.
intros G e t t' Hwf Hof Hof'.
so (print_complete _#3 Hwf Hof) as (u & Hprint & Heqc).
so (print_complete _#3 Hwf Hof') as (u' & Hprint' & Heqc').
so (print_fun _#4 Hprint Hprint'); subst u'.
eauto using eqc_symm, eqc_trans.
Qed.
